(*-------------------------------------------------------------------------*)
(*                             NEFCLASS-PC                                 *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Crisp classification by a fuzzy perceptron with fuzzy antecedents and   *)
(* crisp conclusion weights                                                *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Filename: NEFCLASS.PAS                                                  *)
(* Programmer: Ulrike Nauck                                                *)
(* Last changes: 11/12/95                                                  *)
(*-------------------------------------------------------------------------*)
(* (c) Technical University of Braunschweig, 1994                          *)
(*-------------------------------------------------------------------------*)

program NEFCLASS;

uses Dos, Memory, NFCL_Con, Drivers, Objects, Views, Menus, App, Dialogs,
  Editors, StdDlg, Validate, MyValid, MsgBox, TextView, HelpFile,
  NFCL_Dlg, NFCL_Typ, NFCL_Tsk, NFCL_Obj, NFCL_Run, NFCLHELP;

type

 NFCL_App = object (TApplication)
     ClipboardWindow     : PEditWindow;
     LearnParamDialog    : PLearnParamDialog;
     LearnParamBuffer    : PRLearnParamBuf;
     LearnParam          : PRLearnParam;
     ControlParamDialog  : PControlParamDialog;
     ControlParamBuffer  : PRControlParamBuf;
     ControlParam        : PRControlParam;
     NetParamDialog      : PNetParamDialog;
     NetParamBuffer      : PRNetParam;
     RuleDialog          : PRuleDialog;
     RuleBuffer          : PRRuleParam;
     NewNet              : Boolean;
     LTask               : PTask;
     Net                 : PNefclass;
     LStats              : PStatTask;
     MemStart            : LongInt;
     constructor Init;
     destructor Done; virtual;
     function GetPalette: PPalette; virtual;
     procedure DoAboutBox;
     procedure DoLicenceBox;
     procedure DoHeapBox;
     procedure Learning;
     procedure Testing;
     procedure Statistics;
     procedure OpenLearnParamDialog;
     procedure SetLearnParams;
     procedure GetLearnParams;
     procedure CompleteLearnParams;
     procedure OpenControlParamDialog;
     procedure SetControlParams;
     procedure GetControlParams;
     function  OpenNetParamDialog (HelpContext:integer): Boolean;
     procedure SetNetParams;
     procedure NewNetwork;
     procedure ChangeNetwork;
     procedure LoadNetwork;
     procedure SaveNetwork (Command: Word);
     procedure WriteNetRules;
     procedure SetRuleBuffer;
     procedure GetRuleBuffer;
     procedure DisposeRuleBuffer;
     procedure EditNetRules;
     procedure ShowNetParams;
     procedure OpenNetworkLabelDialog;
     procedure LoadPattern;
     procedure DisplayPattern;
     procedure GetEvent(var Event: TEvent); virtual;
     function  CheckClose: Boolean;
     procedure HandleEvent(var Event: TEvent); virtual;
     procedure InitMenuBar; virtual;
     procedure InitStatusLine; virtual;
     procedure LoadDesktop;
     procedure NewWindow;
     procedure OpenWindow;
     procedure ChDirDialog;
     procedure SaveDesktop;
 end;

var
 NFC           : NFCL_App;
 NetFile       : Text;
 PatternFile   : Text;
 RuleFile      : Text;

(*-------------------------------------------------------------------------*)
(*                        Methods of NFCL_App                              *)
(*-------------------------------------------------------------------------*)

procedure NFCLStreamError(var S: TStream); far;
var
  ErrorMessage: String;
begin
  case S.Status of
    stError: ErrorMessage := 'Stream access error';
    stInitError: ErrorMessage := 'Cannot initialize stream';
    stReadError: ErrorMessage := 'Read beyond end of stream';
    stWriteError: ErrorMessage := 'Cannot expand stream';
    stGetError: ErrorMessage := 'Unregistered type read from stream';
    stPutError: ErrorMessage := 'Unregistered type written to stream';
    end;
  DoneVideo;
  PrintStr('Error: ' + ErrorMessage);
  Halt(Abs(S.Status));
end;

(*-------------------------------------------------------------------------*)

constructor NFCL_App.Init;
var
  R: TRect;
begin
  MaxHeapSize := 8192;
  EditorDialog := StdEditorDialog;
  StreamError := @NFCLStreamError;
  RegisterMenus;
  RegisterObjects;
  RegisterViews;
  RegisterApp;
  RegisterEditors;
  RegisterHelpFile;
  RegisterDialogs;
  RegisterValidate;
  RegisterMyValidate;
  RegisterMyDialogs;
  ResFile.Init(New(PBufStream, Init('NEFCLASS.TVR', stOpenRead, 1024)));
  inherited Init;
  Desktop^.GetExtent(R);
  ClipboardWindow := New(PEditWindow, Init(R, '', wnNoNumber));
  if ValidView(ClipboardWindow) <> nil then
  begin
    ClipboardWindow^.Hide;
    InsertWindow(ClipboardWindow);
    Clipboard := ClipboardWindow^.Editor;
    Clipboard^.CanUndo := False;
  end;
  LearnParam := nil;
  ControlParam := nil;
  RuleBuffer := nil;
  NewNet     := True;
  LTask      := nil;
  Net        := nil;
  LStats      := nil;
  SetControlParams;
  GetControlParams;
  SetLearnParams;
  GetLearnParams;
  CompleteLearnParams;
  DoAboutBox;
  DoLicenceBox;
  MemStart := MemAvail;
end;

(*-------------------------------------------------------------------------*)

destructor NFCL_App.Done;
begin
  if LearnParamBuffer   <> nil then Dispose (LearnParamBuffer);
  if ControlParamBuffer <> nil then Dispose (ControlParamBuffer);
  if NetParamBuffer     <> nil
    then begin
     if NetParamBuffer^.ListPtr <> nil
       then Dispose (NetParamBuffer^.ListPtr, Done);
     if NetParamBuffer^.CListPtr <> nil
       then Dispose (NetParamBuffer^.CListPtr, Done);
     Dispose (NetParamBuffer);
    end;
  DisposeRuleBuffer;
  if LearnParam         <> nil then Dispose (LearnParam);
  if LTask              <> nil then Dispose (LTask, Done);
  if Net                <> nil then Dispose (Net, Done);
  if LStats             <> nil then Dispose (LStats, Done);
  ResFile.Done;
  inherited Done;
end;

(*-------------------------------------------------------------------------*)

function NFCL_App.GetPalette: PPalette;
const
  CNewColor = CAppColor + CHelpColor;
  CNewBlackWhite = CAppBlackWhite + CHelpBlackWhite;
  CNewMonochrome = CAppMonochrome + CHelpMonochrome;
  P: array[apColor..apMonochrome] of string[Length(CNewColor)] =
    (CNewColor, CNewBlackWhite, CNewMonochrome);
begin
  GetPalette := @P[AppPalette];
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.OpenLearnParamDialog;
 begin
  LearnParamDialog := PLearnParamDialog(ResFile.Get('PARAMETER'));
  LearnParamDialog^.HelpCTX := hcParamDialog;
  if LearnParamBuffer = nil then SetLearnParams;
  if ExecuteDialog (LearnParamDialog, LearnParamBuffer) <> cmCancel
    then GetLearnParams;
 end; (* NFCL_App.OpenLearnParamDialog *)

(*-------------------------------------------------------------------------*)
(**)
procedure NFCL_App.GetLearnParams;
 var
  Code: integer;
 begin
  if LearnParam = nil then new (LearnParam);
  with LearnParamBuffer^, LearnParam^ do begin
    nopassing := NoPass = 0;
    intersect05 := Intersect = 0;
    w_learn    := FixedConc = 1;
    free_w_learn := UnitConc = 1;
    asymmetric := Asymmetrical = 0;
    RuleLearning := RuleLearningType(RuleCreation);
    ReLearnRules := ReLearn = 0;
    val (sigma_a, sa, Code);
    val (sigma_b, sb, Code);
    val (sigma_c, sc, Code);
    val (sigma,   s, Code);
  end;(* with *)
 end; (* NFCL_App.GetLearnParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.CompleteLearnParams;
 var
  Code: integer;
 begin
  with LearnParam^ do begin
   EndEpoch     := ControlParam^.EndEpoch;
   ErrEpochs    := ControlParam^.ErrEpochs;
   UpdateRate   := ControlParam^.UpdateRate;
   Misclass_OK  := ControlParam^.Misclass_OK;
   StopError    := ControlParam^.StopError;
   StopCheck    := ControlParam^.StopCheck;
   TextDisplay  := ControlParam^.TextDisplay;
   single_step  := ControlParam^.single_step;
  end;(* with *)
 end; (* NFCL_App.CompleteLearnParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SetLearnParams;
 begin
  if LearnParamBuffer = nil then New (LearnParamBuffer);
  with LearnParamBuffer^ do begin
    NoPass := 0;
    Asymmetrical := 0;
    Intersect:= 1;
    FixedConc:= 0;
    UnitConc := 0;
    RuleCreation := 1;
    ReLearn := 1;
    sigma_a  := '0.01';
    sigma_b  := '0.01';
    sigma_c  := '0.01';
    sigma    := '0.01';
  end; (* with *)
 end; (* NFCL_App.SetLearnParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.OpenControlParamDialog;
 var
  C : Word;
 begin
  ControlParamDialog := PControlParamDialog(ResFile.Get('CONTROL'));
  ControlParamDialog^.HelpCTX := hcControlDialog;
  if ControlParamBuffer = nil then SetControlParams;
  if ExecuteDialog (ControlParamDialog, ControlParamBuffer) <> cmCancel
    then GetControlParams;
 end; (* NFCL_App.OpenLearnParamDialog *)


(*-------------------------------------------------------------------------*)
procedure NFCL_App.GetControlParams;
 var
  Code: integer;
 begin
  if ControlParam = nil then new (ControlParam);
  with ControlParamBuffer^, ControlParam^ do begin
    val (Epochs, EndEpoch, Code);
    val (StopErr, StopError, Code);
    val (NofErrors, MisClass_OK, Code);
    val (Update, UpdateRate, Code);
    TextDisplay := True;
    single_step := false;
    StopCheck := StopCriterion <> 0;
    val (ErrorEpochs, ErrEpochs, Code);
  end;(* with *)
 end; (* NFCL_App.GetControlParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SetControlParams;
 begin
  if ControlParamBuffer = nil then New (ControlParamBuffer);
  with ControlParamBuffer^ do begin
    Epochs   := '500';
    StopErr  := '0.0';
    NofErrors:= '0';
    StopCriterion := 0;
    ErrorEpochs := '100';
    Update   := '10';
  end; (* with *)
 end; (* NFCL_App.SetControlParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.Learning;
 var
   Trainer: PTrainer;
 begin
  if (Net <> nil) and (LTask <> nil)
    then begin
     if (Net^.L1 > 0) and (LTask^.count > 0) and (LTask^.ClassInfo)
       then begin
        CompleteLearnParams;
        Trainer := New(PTrainer, Init (Net, LTask, LearnParam));
        Trainer^.HelpCTX := hcLearnDialog;
        if ValidView(Trainer) <> nil
          then begin
           InsertWindow (Trainer);
           Trainer^.Training (NewNet);
           NewNet := False;
          end;
       end
       else if not LTask^.ClassInfo
              then MessageBox (#3'There is no class information!'#13#3+
                               'These patterns are not for training!', nil,
                               mfError or mfOKButton);
    end
    else begin
     if (Net = nil) and (LTask = nil)
       then MessageBox (#3'You must first define a network'#13#3+
                          'and load a set of patterns!', nil,
                           mfError or mfOKButton)
       else if Net = nil
              then MessageBox (#3'You must define a network'#13#3+
                                 'before training!', nil,
                                 mfError or mfOKButton)
              else MessageBox (#3'You must load a set of patterns'#13#3+
                                 'before training!', nil,
                                 mfError or mfOKButton)
    end;
 end; (* NFCL_App.Learning *)

(*-------------------------------------------------------------------------*)
procedure NFCL_App.Testing;
 var
   Tester: PTester;
 begin
  if (Net <> nil) and (LTask <> nil)
    then begin
     if (Net^.L2 > 0) and (LTask^.count > 0)
       then begin
        Tester := New(PTester, Init (Net, LTask));
        Tester^.HelpCTX := hcTestDialog;
        if ValidView(Tester) <> nil
          then begin
           InsertWindow (Tester);
          end;
       end
       else if Net^.L2 <= 0
              then MessageBox (#3'There are no rules in the net!'#13#3+
                               'Testing patterns is not possible!', nil,
                        mfError or mfOKButton)
    end
    else MessageBox (#3'You must first define a network'#13#3+
                       'and patterns!', nil,
                     mfError or mfOKButton);
 end; (* NFCL_App.Testing *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.Statistics;
 var
  Stats: PStatistics;
 begin
  if (Net <> nil) and (LTask <> nil)
    then begin
     if (Net^.L1 > 0) and (LTask^.count > 0)
       then begin
        if LStats = nil
          then LStats := New (PStatTask, Init (LTask));
        Stats := New(PStatistics, Init (Net, LTask, LStats, NewNet));
        Stats^.HelpCTX := hcStatisticDialog;
        if ValidView(Stats) <> nil
          then InsertWindow (Stats);
       end
    end
    else MessageBox (#3'You must first define a network'#13#3+
                       'and patterns!', nil,
                     mfError or mfOKButton);
 end; (* NFCL_App.Statistics *)

(*-------------------------------------------------------------------------*)


function NFCL_App.OpenNetParamDialog (HelpContext:integer): Boolean;
 var
   i          : integer; (* loop variable *)
   Code       : integer; (* code for interpreting string as value *)
   L1, L3     : integer; (* number of units in input and output layer *)
   nofs       : integer; (* number of fuzzy sets per unit *)
   maxrules   : integer; (* maximum number of units in rule layer *)
   OK         : boolean; (* true, if new parameters entered *)
   UseSum     : boolean; (* true, if aggregation by weighted sum *)
   Individual : boolean; (* true, if individual number of fuzzy sets *)
   Result     : word;    (* Result of ExecuteDialog (button pressed) *)
   S          : string[3];

 begin
  OK := False;
  NetParamDialog := PNetParamDialog(ResFile.Get('NETPARAM'));
  NetParamDialog^.HelpCTX := HelpContext;
  if NetParamBuffer = nil then SetNetParams;
  if HelpContext = hcNetDisplay
    then begin
     Str (Net^.L2, S);
     NetParamBuffer^.RuleLayer := S + ' (' + NetParamBuffer^.RuleLayer + ')';
     with NetParamDialog^ do
      begin (* switch of all input areas of the dialog window *)
       InLayer^.Options := InLayer^.Options
        and not ofSelectable and not ofFirstClick;
       RuleLayer^.Options := RuleLayer^.Options
        and not ofSelectable and not ofFirstClick;
       RuleLayer^.SetValidator (nil);
       OutLayer^.Options := OutLayer^.Options
        and not ofSelectable and not ofFirstClick;
       UseSum^.EnableMask := 0;
       Individual^.EnableMask := 0;
       Partition^.Options := Partition^.Options
        and not ofSelectable and not ofFirstClick;
       Picked^.Options := Picked^.Options
        and not ofSelectable and not ofFirstClick;
       IndFSets^.Options := IndFSets^.Options
        and not ofSelectable and not ofFirstClick;
       ClassPicked^.Options := ClassPicked^.Options
        and not ofSelectable and not ofFirstClick;
      end; (* with *)
    end;
  Result := ExecuteDialog (NetParamDialog, NetParamBuffer);
  if  (Result = cmOK) and (HelpContext <> hcNetDisplay)
    then begin
     val (NetParamBuffer^.InLayer, L1, Code);
     val (NetParamBuffer^.RuleLayer, maxrules, Code);
     val (NetParamBuffer^.OutLayer, L3, Code);
     val (NetParamBuffer^.Partition, nofs, Code);
     UseSum := NetParamBuffer^.UseSum = 0;
     Individual := NetParamBuffer^.Individual = 1;
     if Net <> nil
       then begin
        Dispose (Net, Done);
        Net := nil;
       end;
     New (Net, Init(L1, maxrules, L3, nofs, NetParamBuffer^.ListPtr,
                    NetParamBuffer^.CListPtr, Individual, UseSum));
     OK := True;
    end;
  (* Return True, if new net was defined, or if weights must be displayed *)
  OpenNetParamDialog := OK or (Result = cmYes);
 end; (* NFCL_App.OpenNetParamDialog *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SetNetParams;
 var
  APartition : PPartition;
  i          : integer;
  S          : string[3];
  CStr       : PVNStr;
 begin
  if NetParamBuffer = nil
   then begin
    New (NetParamBuffer);
    NetParamBuffer^.ListPtr := nil;
    NetParamBuffer^.CListPtr := nil;
   end;
  with NetParamBuffer^ do begin
   if Net <> nil
     then begin
      str (Net^.L1, InLayer);
      if NewNet
        then str (Net^.max_rules, RuleLayer)
        else str (Net^.max_rules, RuleLayer);
      str (Net^.L3, OutLayer);
      str (Net^.Antecbase[1].count, Partition);
      if Net^.UseSum
        then UseSum := 0
        else UseSum := 1;
      if Net^.Individual
        then Individual := 1
        else Individual := 0;

      if NetParamBuffer^.ListPtr = nil
        then begin
         New (NetParamBuffer^.ListPtr, Init (maxunits, 1));
         for i := 1 to maxunits do
          begin
           new (APartition);
           APartition^.Count := 3;
           Str (i, S);
           APartition^.VName := 'V_'+S;
           NetParamBuffer^.ListPtr^.Insert (APartition);
          end;
        end;

      if NetParamBuffer^.CListPtr = nil
        then begin
         New (NetParamBuffer^.CListPtr, Init (maxunits, 1));
         for i := 1 to maxunits do
          begin
           New (CStr);
           Str (i, S);
           CStr^ := 'Class_' + S;
           NetParamBuffer^.CListPtr^.Insert (CStr);
          end;
        end;

      for i := 1 to Net^.L1 do
       begin
        PPartition(NetParamBuffer^.ListPtr^.At(i-1))^.VName :=
         Net^.Antecbase[i].VName;
        PPartition(NetParamBuffer^.ListPtr^.At(i-1))^.Count :=
         Net^.Antecbase[i].Count;
       end;

      for i := 1 to Net^.L3 do
        PVNStr(NetParamBuffer^.CListPtr^.At(i-1))^ := Net^.ClassNames[i];

      Picked := Net^.AntecBase[1].VName;
      Str (Net^.AntecBase[1].Count, IndFSets);
      Focussed := 0;
      CFocussed := 0;
      ClassPicked := Net^.ClassNames[1];
     end (* if Net <> nil then *)
     else begin
      InLayer   := '4';
      RuleLayer := '7';
      OutLayer  := '3';
      Partition := '3';
      UseSum    := 0;
      Individual:= 1;
      Picked := 'V_1';
      IndFSets := '3';
      Focussed := 0;
      CFocussed := 0;
      ClassPicked := 'Class_1';
      if NetParamBuffer^.ListPtr <> nil
        then Dispose (NetParamBuffer^.ListPtr, Done);
      if NetParamBuffer^.CListPtr <> nil
        then Dispose (NetParamBuffer^.CListPtr, Done);
      New (NetParamBuffer^.ListPtr, Init (maxunits, 1));
      New (NetParamBuffer^.CListPtr, Init (maxunits, 1));
      for i := 1 to maxunits do
       begin
        new (APartition);
        APartition^.Count := 3;
        Str (i, S);
        APartition^.VName := 'V_'+S;
        NetParamBuffer^.ListPtr^.Insert (APartition);
        New (CStr);
        CStr^ := 'Class_' + S;
        NetParamBuffer^.CListPtr^.Insert (CStr);
       end;
     end;
  end;
 end; (* NFCL_App.SetNetParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.NewNetwork;
 begin
  if Net <> nil
    then begin
     if MessageBox (#3'There is already a network defined!'#13#3+
                        'This network will be deleted. If'#13#3+
                        'there are patterns they will also'#13#3+
                        'be deleted. OK to proceed?', nil,
                        mfInformation or mfOKButton or mfCancelButton)
        <> cmCancel
       then begin
         SetNetParams;
         Dispose (Net, Done);
         DisposeRuleBuffer;
         Net := nil;
         if LTask <> nil
           then begin
            Dispose (LTask, Done);
            LTask := nil;
           end;
         if LStats <> nil
           then begin
            Dispose (LStats, Done);
            LStats := nil;
           end;
         DisableCommands([cmYes]);
         NewNet := OpenNetParamDialog (hcNetNew);
         EnableCommands([cmYes]);
       end;
    end
    else begin
     SetNetParams;
     DisableCommands([cmYes]);
     NewNet := OpenNetParamDialog (hcNetNew);
     EnableCommands([cmYes]);
    end;
 end; (* NFCL_App.NewNetwork *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.ChangeNetwork;
 var
   Code       : integer; (* code for interpreting string as value *)
   nofs       : integer; (* number of fuzzy sets per unit *)
   maxrules   : integer; (* maximum number of units in rule layer *)
   UseSum     : boolean; (* true, if aggregation by weighted sum *)
   Individual : boolean; (* true, if individual number of fuzzy sets *)
   Result     : word;    (* Result of ExecuteDialog (button pressed) *)
 begin
  if Net = nil
    then MessageBox (#3'There is no network defined!', nil,
                      mfError or mfOKButton)
    else begin
     if not Net^.Saved
       then Result :=
            MessageBox (#3'The network is not saved! Changing the network '+
                        'can e.g. lead to deletion of rules. Continue '+
                        'anyway?', nil,
                        mfWarning or mfYesButton or mfNoButton)
       else Result := cmYes;
     if Result = cmYes
      then begin
       SetNetParams;
       NetParamDialog := PNetParamDialog(ResFile.Get('NETPARAM'));
       NetParamDialog^.HelpCTX := hcNetChange;
       NetParamDialog^.InLayer^.Options :=
         NetParamDialog^.InLayer^.Options and not ofSelectable
                                          and not ofFirstClick;
       NetParamDialog^.OutLayer^.Options :=
         NetParamDialog^.OutLayer^.Options and not ofSelectable
                                          and not ofFirstClick;
       DisableCommands([cmYes]);
       Result := ExecuteDialog (NetParamDialog, NetParamBuffer);
       EnableCommands([cmYes]);
       if  Result = cmOK
         then begin
          val (NetParamBuffer^.RuleLayer, maxrules, Code);
          val (NetParamBuffer^.Partition, nofs, Code);
          UseSum := NetParamBuffer^.UseSum = 0;
          Individual := NetParamBuffer^.Individual = 1;
          Net^.ChangeNet(maxrules, nofs, NetParamBuffer^.ListPtr,
                         NetParamBuffer^.CListPtr, Individual, UseSum);
         end;
      end; (* if Result = cmYes then *)
    end; (* else *)
 end; (* NFCL_App.ChangeNetwork *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.ShowNetParams;
 var
  OK : Boolean;
 begin
  if Net = nil
    then MessageBox (#3'There is no network defined!', nil,
                      mfError or mfOKButton)
    else begin
     SetNetParams;
     DisableCommands ([cmFSet, cmCSet]);
     OK := OpenNetParamDialog (hcNetDisplay);
     EnableCommands ([cmFSet, cmCSet]);
     if OK then Net^.DisplayFuzzyWeights;
    end;
 end; (* NFCL_App.ShowNetParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.LoadNetwork;
var
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
  OK        : Boolean;
const
  FDOptions: Word = fdOKButton or fdOpenButton or fdHelpButton;
begin
  if Net <> nil
    then OK := MessageBox (#3'There is already a network defined!'#13#3+
                           'This network will be deleted. If'#13#3+
                           'there are patterns they will also'#13#3+
                           'be deleted. OK to proceed?', nil,
                           mfInformation or mfOKButton or mfCancelButton)
               <> cmCancel
     else OK := True;
  if OK
    then begin
     if Net <> nil
       then begin
        Dispose (Net, Done);
        DisposeRuleBuffer;
        Net := nil;
        if LTask <> nil
          then begin
           Dispose (LTask, Done);
           LTask := nil;
          end;
        if LStats <> nil
          then begin
           Dispose (LStats, Done);
           LStats := nil;
          end;
       end;
     TheFile := '*.NET';
     New(FileDialog, Init(TheFile, 'Open network file', '~F~ile name',
       FDOptions, 1));
     FileDialog^.HelpCTX := hcNetLoad;
     if ExecuteDialog(FileDialog, @TheFile) <> cmCancel
       then begin
        if Filename (NetFile, reading, TheFile)
          then begin
           New (Net, Init(0,0,0,0,nil,nil,false,true));
           if Net^.ReadNet (NetFile, TheFile)
                 then begin
                   MessageBox ('Network file was read successfully', nil,
                                mfInformation or mfOKButton);
                   if Net^.InVarRange[1].min > Net^.InVarRange[1].max
                     then NewNet := True (* uninitialized net was loaded *)
                     else NewNet := False;
                   if Net^.L2 = 0 (* loaded net has no rules yet *)
                     then Net^.LearnRules := True;
                   SetNetParams;
                 end
                 else begin
                  Dispose (Net, Done);
                  DisposeRuleBuffer;
                  Net := nil;
                  MessageBox ('Could not read Network file', nil,
                              mfError or mfOKButton);
                 end;
           end; (* if Net^.ReadNet then *)
       end; (* if Filename then *)
    end; (* if ExecuteDialog then *)
end; (* NFCL_App.LoadNetwork *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SaveNetwork (Command: Word);
var
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
  OK        : Boolean;
const
  FDOptions: Word = fdOKButton or fdHelpButton;
begin
  if Net = nil
    then MessageBox (#3'There is no network defined!', nil,
                     mfError or mfOKButton)
    else begin
     if (command = cmNetworkSaveAs) or (Net^.Name = '')
       then begin
        TheFile := '*.NET';
        New(FileDialog, Init(TheFile, 'Save network file as', '~F~ile name',
          FDOptions, 1));
        FileDialog^.HelpCTX := hcNetSave;
        if Net^.Name <> '' then TheFile := Net^.Name;
        OK := ExecuteDialog(FileDialog, @TheFile) <> cmCancel;
       end
       else begin
        OK := True;
        TheFile := Net^.Name;
       end;
     if OK
      then begin
       if Filename (NetFile, writing, TheFile)
         then begin
          Net^.WriteNet (NetFile, TheFile, Net^.NetLabel);
          MessageBox ('Network file was written successfully', nil,
                      mfInformation or mfOKButton)
         end
         else MessageBox ('Could not write Network file', nil,
                          mfError or mfOKButton);
      end; (* if OK then *)
    end; (* if Net = nil ... else *)
end; (* NFCL_App.SaveNetwork *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.WriteNetRules;
var
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
const
  FDOptions: Word = fdOKButton or fdHelpButton;
begin
  if Net = nil
    then MessageBox (#3'There is no network defined!', nil,
                     mfError or mfOKButton)
    else begin
     if Net^.L2 <= 0
       then MessageBox (#3'The network has no rules yet!', nil,
                     mfError or mfOKButton)
       else begin
        TheFile := '*.RUL';
        New(FileDialog, Init(TheFile, 'Save rules in file', '~F~ile name',
          FDOptions, 1));
        FileDialog^.HelpCTX := hcNetRules;
        if Net^.FileName <> '' then TheFile := Net^.FileName + '.RUL';
        if ExecuteDialog(FileDialog, @TheFile) <> cmCancel then
        begin
         if Filename (RuleFile, writing, TheFile)
           then begin
            Rewrite (RuleFile);
            Net^.WriteRuleBase (RuleFile);
            Close (RuleFile);
            MessageBox ('Rule base was written successfully!', nil,
                        mfInformation or mfOKButton)
           end
           else MessageBox ('Could not write rule base file!', nil,
                            mfError or mfOKButton);
       end; (* else *)
    end; (* else *)
  end;
end; (* NFCL_App.WriteNetRules *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SetRuleBuffer;
 var
  Name: PVNStr;
  Rule: PRuleDescriptor;
  Antec: PAntStr;
  S   : string[3];
  i,j   : integer;

 begin

  if RuleBuffer = nil
    then begin
     new(RuleBuffer);
     RuleBuffer^.RListPtr := nil;
     RuleBuffer^.CListPtr := nil;
     RuleBuffer^.AListPtr := nil;
     RuleBuffer^.VListPtr := nil;
     RuleBuffer^.FListPtr := nil;
    end;

  if RuleBuffer^.RListPtr = nil
    then begin
     New (RuleBuffer^.RListPtr, Init (Net^.max_rules, 1));
     for i := 1 to Net^.L2 do
      begin
       New (Rule);
       Str (i, S);
       Rule^.Name := 'R_' + S;
       Rule^.Conclusion := Net^.ClassNames[Net^.ConclusionWeights[i].node];
       for j := 1 to Net^.L1 do
        begin
         if Net^.Antecedents[j,i] > 0
          then begin
           Rule^.Antecedent[j].Variable := Net^.AntecBase[j].VName;
           Rule^.Antecedent[j].Value :=
               Net^.AntecBase[j].FSets[Net^.Antecedents[j,i]].Name;
          end
          else begin
           Rule^.Antecedent[j].Variable := '';
           Rule^.Antecedent[j].Value := '';
          end;
        end;
       RuleBuffer^.RListPtr^.Insert (Rule);
      end;
    end;

  if RuleBuffer^.CListPtr = nil
    then begin
     New (RuleBuffer^.CListPtr, Init (Net^.L3, 1));
     for i := 1 to Net^.L3 do
      begin
       New (Name);
       Name^ := Net^.ClassNames[i];
       RuleBuffer^.CListPtr^.Insert (Name);
      end;
    end;

  if RuleBuffer^.VListPtr = nil
    then begin
     New (RuleBuffer^.VListPtr, Init (Net^.L1, 1));
     for i := 1 to Net^.L1 do
      begin
       New (Name);
       Name^ := Net^.AntecBase[i].VName;
       RuleBuffer^.VListPtr^.Insert (Name);
      end;
    end;

  if RuleBuffer^.FListPtr = nil
    then begin
     New (RuleBuffer^.FListPtr, Init (maxfs, 1));
    end;

  if RuleBuffer^.AListPtr = nil
    then begin
     New (RuleBuffer^.AListPtr, Init (Net^.L1, 1));
    end;

  RuleBuffer^.CFocussed := 0;
  RuleBuffer^.AFocussed := 0;
  RuleBuffer^.RFocussed := 0;
  RuleBuffer^.FFocussed := 0;
  RuleBuffer^.VFocussed := 0;

  if Net^.L2 > 0
    then RuleBuffer^.Class := Net^.ClassNames[Net^.ConclusionWeights[1].Node]
    else RuleBuffer^.Class := '';

 end; (* NFCL_App.SetRuleBuffer *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.GetRuleBuffer;

 var
  ARule : PRuleDescriptor;
  i,j   : integer;

 function FindFS (Variable: integer; AName: FSNStr): byte;
  var
   i : integer;
  begin
   if AName = ''
     then FindFS := 0
     else begin
      i := 0;
      repeat
       inc (i)
      until (AName = Net^.AntecBase[Variable].FSets[i].Name) or
            (i >= Net^.AntecBase[Variable].count);
      FindFS := i;
     end;
  end; (* FindFS *)

  function FindClass (AClass: VNStr): integer;
   var
    i : integer;
   begin
    i := 0;
    repeat
     inc (i)
    until (AClass = Net^.ClassNames[i]) or
          (i >= Net^.L3);
    FindClass := i;
   end; (* FindClass *)

 begin
  Net^.L2 := RuleBuffer^.RListPtr^.Count;
  for i := 1 to Net^.L2 do
   begin
    ARule := PRuleDescriptor(RuleBuffer^.RListPtr^.At(i-1));
    for j := 1 to Net^.L1
     do Net^.Antecedents[j,i] := FindFS (j, ARule^.Antecedent[j].Value);
    Net^.ConclusionWeights[i].Node := FindClass (ARule^.Conclusion);
    Net^.ConclusionWeights[i].Weight := 1.0;
   end;
  DisposeRuleBuffer;
 end; (* NFCL_App.GetRuleBuffer *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.DisposeRuleBuffer;
 begin
  if RuleBuffer <> nil
    then begin
     with RuleBuffer^ do begin
      if AListPtr <> nil then Dispose (AListPtr, Done);
      if CListPtr <> nil then Dispose (CListPtr, Done);
      if FListPtr <> nil then Dispose (FListPtr, Done);
      if RListPtr <> nil then Dispose (RListPtr, Done);
      if VListPtr <> nil then Dispose (VListPtr, Done);
     end;
     Dispose (RuleBuffer);
     RuleBuffer := nil;
    end;
 end; (* NFCL_App.DisposeRuleBuffer *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.EditNetRules;
 begin
  if Net = nil
    then MessageBox (#3'You must first define a network!', nil,
                     mfError or mfOKButton)
    else begin
     SetRuleBuffer;
     RuleDialog := PRuleDialog(ResFile.Get('RULES'));
     RuleDialog^.InitData (Net);
     if ExecuteDialog (RuleDialog, RuleBuffer) <> cmCancel
       then GetRuleBuffer
       else DisposeRuleBuffer;
    end;
 end; (* NFCL_App.EditNetRules *)

(*-------------------------------------------------------------------------*)
 procedure NFCL_App.OpenNetworkLabelDialog;
  var
   ALabel: LabelStr;
  begin
   if Net = nil
     then MessageBox (#3'You must first define a network!', nil,
                     mfError or mfOKButton)
     else begin
      ALabel := Net^.NetLabel;
      if ExecuteDialog (PNetLabelDialog (ResFile.Get('NETLABEL')),
                        @ALabel) <> cmCancel
        then Net^.NetLabel := ALabel;
     end;
  end; (* NFCL_App.OpenNetworkLabelDialog *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.LoadPattern;
var
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
  OK        : Boolean;
  Path      : PathStr;
  Dir       : DirStr;
  Name      : NameStr;
  Ext       : ExtStr;
const
  FDOptions: Word = fdOKButton or fdOpenButton or fdHelpButton;
begin
  OK := True;
  if LTask = nil then new(LTask, Init);
  if LTask^.Count > 0
    then begin
      OK := (MessageBox (#3'There are already patterns defined!'#13#3+
                         'These patterns will be deleted.'#13#3+
                         'OK to proceed?', nil,
                         mfInformation or mfOKButton or mfCancelButton)
             <> cmCancel);
      if OK
        then begin
         Dispose (LTask, Done); LTask := nil;
         New (LTask, Init);
        end;
    end;
  if Net = nil
    then begin
     OK := False;
     MessageBox (#3'You must first define a network!', nil,
                 mfError or mfOKButton);
     Dispose (LTask, Done); LTask := nil;
    end;
  if OK
    then begin
     TheFile := '*.dat';
     New(FileDialog, Init(TheFile, 'Open pattern file', '~F~ile name',
       FDOptions, 1));
     FileDialog^.HelpCTX := hcPatLoad;
     if ExecuteDialog(FileDialog, @TheFile) <> cmCancel
       then begin
        if Filename (PatternFile, reading, TheFile)
          then if LTask^.ReadPattern (PatternFile, Net^.L1, Net^.L3)
                 then begin
                   MessageBox (#3'Pattern file was read successfully', nil,
                                mfInformation or mfOKButton);
                   FSplit(TheFile, Dir, Name, Ext);
                   if LTask^.TaskName = ''
                     then LTask^.TaskName := Name+Ext;
                   LTask^.FileName := Name;
                   if LStats <> nil
                     then begin
                      Dispose (LStats, Done);
                      LStats := nil;
                     end;
                 end
                 else begin
                   MessageBox (#3'Pattern file does not fit'#13#3+
                                  'to network',
                                   nil, mfError or mfOKButton);
                   Dispose (LTask, Done); LTask := nil;
                   if LStats <> nil
                     then begin
                      Dispose (LStats, Done);
                      LStats := nil;
                     end;
                 end;
       end (* if ExecuteDialog then *)
       else begin (* reading the patterns was canceled *)
        Dispose (LTask, Done);
        LTask := nil;
        if LStats <> nil
          then begin
           Dispose (LStats, Done);
           LStats := nil;
          end;
       end;
    end;
end; (* NFCL_App.LoadPattern *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.DisplayPattern;
 var
  TermWin     : PTermWin;
  Term        : PText;
  i           : integer;
 begin
  if LTask <> nil
    then begin
     New(TermWin, Init ('Parameters of loaded patterns', 41, 15));
     if ValidView(TermWin) <> nil
       then begin
        InsertWindow (TermWin);
        TermWin^.GetTerm (Term);
        writeln (Term^, 'Name of Pattern Set:');
        writeln (Term^, LTask^.TaskName);
        writeln (Term^);
        writeln (Term^, 'Number of Patterns      : ', LTask^.Count);
        writeln (Term^, 'Number of input features: ', LTask^.Nin);
        if LTask^.ClassInfo
          then writeln (Term^, 'Number of classes       : ', LTask^.Nout)
          else writeln (Term^, 'There is no classification information');
        writeln (Term^);
        writeln (Term^, 'Ranges of the input features:');
        for i := 1 to LTask^.Nin do
          writeln (Term^, i:2, ') Min = ', LTask^.min_input[i]:7:2,
                   ', Max = ', LTask^.max_input[i]:7:2);
       end; (* then *)
    end (* then *)
    else
     MessageBox (#3'There are no patterns defined',nil,
                 mfError or mfOkButton);
 end; (* NFCL_App.DisplayPattern *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.DoAboutBox;
begin
  ExecuteDialog(PDialog(ResFile.Get('ABOUTBOX')), nil);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.DoLicenceBox;
begin
  ExecuteDialog(PDialog(ResFile.Get('LICENCEBOX')), nil);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.DoHeapBox;
var
 MaxBlock:  String;
 HeapSum :  String;
 HeapStart: String;
begin
  Str (MaxAvail, MaxBlock);
  Str (MemAvail, HeapSum);
  Str (MemStart, HeapStart);
  MaxBlock := 'Largest Free Block  : ' + MaxBlock + #13;
  HeapSum  := 'Sum of Free Memory  : ' + HeapSum + #13;
  HeapStart:= 'Free Memory at Start: ' + HeapStart;
  MessageBox (MaxBlock + HeapSum + HeapStart, nil,
              mfInformation or mfOkButton);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.GetEvent(var Event: TEvent);
var
  W        : PWindow;
  HFile    : PHelpFile;
  HelpStrm : PDosStream;
const
  HelpInUse: Boolean = False;
begin
  inherited GetEvent(Event);
  case Event.What of
    evCommand:
      if (Event.Command = cmHelp) and not HelpInUse then
      begin
        HelpInUse := True;
        HelpStrm  := New(PDosStream, Init(CalcHelpName, stOpenRead));
        HFile     := New(PHelpFile, Init(HelpStrm));
        if HelpStrm^.Status <> stOk then
        begin
          MessageBox('Could not open help file.', nil, mfError + mfOkButton);
          Dispose(HFile, Done);
        end
        else
        begin
          W := New(PHelpWindow,Init(HFile, GetHelpCtx));
          if ValidView(W) <> nil then
          begin
            ExecView(W);
            Dispose(W, Done);
          end;
          ClearEvent(Event);
        end;
        HelpInUse := False;
      end;
    evMouseDown:
      if Event.Buttons <> 1 then Event.What := evNothing;
  end;
end;

(*-------------------------------------------------------------------------*)

function NFCL_App.CheckClose : Boolean;
 var
  Quit : Word;
 begin
  if Net <> nil
    then begin
      if not Net^.Saved
        then Quit := MessageBox ('The current network was not ' +
                                 'saved yet. '+
                                 'Do you really want to quit?', nil,
                                 mfWarning or mfYesButton or
                                 mfNoButton)
        else Quit := MessageBox (#3'Quit NEFCLASS-PC ?', nil,
                                 mfWarning or mfYesButton or
                                 mfNoButton);
      if Quit = cmYes
       then CheckClose := True
       else CheckClose := False;
    end
    else CheckClose := True;
 end; (* NFCL_App.CheckClose *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.HandleEvent(var Event: TEvent);
var
  R: TRect;
begin
  if Event.What = evCommand then
  begin
    case Event.Command of
      cmQuit:
       begin
        if not CheckClose
          then ClearEvent(Event)
       end;
      cmOptionsLoad:
        begin
          LoadDesktop;
          ClearEvent(Event);
        end;
      cmOptionsSave:
        begin
          SaveDesktop;
          ClearEvent(Event);
        end;
      cmClipShow:
        with ClipboardWindow^ do
        begin
          Select;
          Show;
          ClearEvent(Event);
        end;
      cmNew:
        begin
          NewWindow;
          ClearEvent(Event);
        end;
      cmOpen:
        begin
          OpenWindow;
          ClearEvent(Event);
        end;
      cmChangeDir:
        begin
          ChDirDialog;
          ClearEvent(Event);
        end;
      cmNetworkNew     :
        begin
          NewNetwork;
          ClearEvent(Event);
        end;
      cmNetworkChange  :
        begin
          ChangeNetwork;
          ClearEvent(Event);
        end;
      cmNetworkDisplay :
        begin
          ShowNetParams;
          ClearEvent(Event);
        end;
      cmNetworkLoad    :
        begin
          LoadNetwork;
          ClearEvent(Event);
        end;
      cmNetworkSave    :
        begin
          SaveNetwork (cmNetworkSave);
          ClearEvent(Event);
        end;
      cmNetworkSaveAs  :
        begin
          SaveNetwork (cmNetworkSaveAs);
          ClearEvent(Event);
        end;
      cmNetEditRules   :
        begin
          EditNetRules;
          ClearEvent(Event);
        end;
      cmNetworkRules   :
        begin
          WriteNetRules;
          ClearEvent(Event);
        end;
      cmNetworkLabel   :
        begin
          OpenNetworkLabelDialog;
          ClearEvent(Event);
        end;
      cmPatternLoad    :
        begin
          LoadPattern;
          ClearEvent(Event);
        end;
      cmPatternDisplay :
        begin
          DisplayPattern;
          ClearEvent(Event);
        end;
      cmRunLearn       :
        begin
          Learning;
          ClearEvent(Event);
        end;
      cmRunParam       :
        begin
          OpenLearnParamDialog;
          ClearEvent(Event);
        end;
      cmRunControl     :
        begin
          OpenControlParamDialog;
          ClearEvent(Event);
        end;
      cmRunTest        :
        begin
          Testing;
          ClearEvent(Event);
        end;
      cmRunStats       :
        begin
          Statistics;
          ClearEvent(Event);
        end;
      cmOptionsVideo:
        begin
          SetScreenMode(ScreenMode xor smFont8x8);
          ClearEvent(Event);
        end;
      cmAbout:
        begin
          DoAboutBox;
          ClearEvent(Event);
        end;
      cmLicence:
        begin
          DoLicenceBox;
          ClearEvent(Event);
        end;
      cmHeapSituation:
        begin
         DoHeapBox;
         ClearEvent(Event);
        end;
    end;
  end;
  inherited HandleEvent(Event);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.InitMenuBar;
begin
  MenuBar := PMenuBar(ResFile.Get('MAINMENU'));
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.InitStatusLine;
var
  R: TRect;
begin
  StatusLine := PStatusLine(ResFile.Get('STATUS'));
  GetExtent(R);
  StatusLine^.MoveTo(0, R.B.Y - 1);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.LoadDesktop;
var
  DesktopFile: TBufStream;
  TempDesktop: PDesktop;
  R: TRect;
begin
  DesktopFile.Init('NEFCLASS.DTP', stOpenRead, 1024);
  TempDesktop := PDesktop(DesktopFile.Get);
  DesktopFile.Done;
  if ValidView(TempDesktop) <> nil then
  begin
    Desktop^.Delete(ClipboardWindow);
    Delete(Desktop);
    Dispose(Desktop, Done);
    Desktop := TempDesktop;
    Insert(Desktop);
    GetExtent(R);
    R.Grow(0, -1);
    Desktop^.Locate(R);
    InsertWindow(ClipboardWindow);
  end;
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.NewWindow;
var
  R: TRect;
  TheWindow: PEditWindow;
begin
  R.Assign(0, 0, 60, 20);
  TheWindow := New(PEditWindow, Init(R, '', wnNoNumber));
  InsertWindow(TheWindow);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.OpenWindow;
var
  R: TRect;
  FileDialog: PFileDialog;
  TheFile: FNameStr;
const
  FDOptions: Word = fdOKButton or fdOpenButton or fdHelpButton;
begin
  TheFile := '*.*';
  New(FileDialog, Init(TheFile, 'Open file', '~F~ile name',
    FDOptions, 1));
  FileDialog^.HelpCTX := hcFOFileOpenDBox;
  if ExecuteDialog(FileDialog, @TheFile) <> cmCancel then
  begin
    R.Assign(0, 0, 75, 20);
    InsertWindow(New(PEditWindow, Init(R, TheFile, wnNoNumber)));
  end;
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.ChDirDialog;
var
  CDDialog: PChDirDialog;
const
  CDOptions: Word = cdNormal or cdHelpButton;
begin
  New(CDDialog, Init(CDOptions, 2));
  CDDialog^.HelpCTX := hcFCChDirDBox;
  ExecuteDialog(CDDialog, nil);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SaveDesktop;
var
  DesktopFile: TBufStream;
begin
  Desktop^.Delete(ClipboardWindow);
  DesktopFile.Init('NEFCLASS.DTP', stCreate, 1024);
  DesktopFile.Put(Desktop);
  DesktopFile.Done;
  InsertWindow(ClipboardWindow);
end;

(*-------------------------------------------------------------------------*)

begin
 NFC.Init;
 NFC.Run;
 NFC.Done;
end.