(*-------------------------------------------------------------------------*)
(*                             NEFCLASS-PC                                 *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Crisp classification by a fuzzy perceptron with fuzzy antecedents and   *)
(* crisp conclusion weights                                                *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Filename: NFCL_RUN.PAS, unit for training and testing the network       *)
(* Programmer: Ulrike Nauck                                                *)
(* Last changes: 10/01/95                                                  *)
(*-------------------------------------------------------------------------*)
(* (c) Technical University of Braunschweig, 1994                          *)
(*-------------------------------------------------------------------------*)

Unit NFCL_Run;

interface

uses Dos, Crt, Graph, Drivers, App, Objects, Dialogs, TextView,
     Views, MsgBox, Memory, StdDlg,
     NFCL_Dlg, NFCL_Con, NFCL_Typ, NFCL_Tsk, NFCL_Obj, NFCLHELP, PlotBGI;
type

 PLineBuffer = ^TLineBuffer;
 TLineBuffer = array[0..5*maxunits] of word;

 PTrainer = ^TTrainer;
 TTrainer = object (TDialog)
   Net      : PNefclass;
   LearnTask: PTask;
   TextWin  : PWindow;
   Terminal : PTerminal; (* text window to display learning process *)
   Term     : Text;      (* handle to access the text window *)
   LParam   : PRLearnParam; (* record of learning parameters *)
   from     : integer;   (* first variable to be displayed in graphics *)
   NofInputs: integer;   (* number of input variables of the network *)
   switched : boolean;   (* true, if video mode has changed *)
   waiting  : boolean;   (* true, if waiting for signal in single step mode*)
   continous: boolean;   (* true, if continous update is active in graphics*)
   new_state: boolean;   (* true, if continous/halted/single_step changed*)
   update   : boolean;   (* true, if user wants to update display *)
   halted   : boolean;   (* true, if learning is halted *)
   ForcedStop: boolean;  (* true, if learning is cancelled *)
   LearnEnd  : boolean;  (* true, if learning is over *)
   ErrMsgClose: boolean; (* true, if close was clicked during learning *)

   constructor Init (ANet: PNefclass; ATask: PTask; LP: PRLearnParam);
   destructor Done; virtual;
   procedure HandleEvent (var Event: TEvent); virtual;
   procedure SwitchToGraphics;
   procedure SwitchToText;
   procedure Display_Text_Update (Epoch: integer; Error: real;
                                  misclass, count: integer);
   procedure Training (NewNet: Boolean);
 end;

 PResultWindow = ^TResultWindow;
 TResultWindow = object (TWindow)
   HScrollbar   : PScrollbar;
   OutBuffer    : PLineBuffer;
   TargetBuffer : PLineBuffer;
   constructor Init (ARect: TRect; ARange: integer; OBuf, TBuf: PLineBuffer);
   destructor Done; virtual;
   procedure HandleEvent (var Event: TEvent); virtual;
   procedure CopyLine (Buf: PLineBuffer; var ResBuf: TLineBuffer;
                       Start, Stop, Chunk: integer);
   procedure Draw; virtual;
 end;

 PTester = ^TTester;
 TTester = object (TDialog)
   PatternNumber: PInputLine;
   ErrorNumber  : PInputLine;
   PatternError : PInputLine;
   MeanError    : PInputLine;
   Result       : PCluster;

   Task       : PTask;
   Net        : PNefclass;
   TextWin    : PResultWindow; (* window to display propagation result *)
   patno      : integer;
   OutBuf     : PLineBuffer;
   TargetBuf  : PLineBuffer;
   constructor Init (ANet: PNefclass; ATask: PTask);
   destructor Done; virtual;
   procedure HandleEvent (var Event: TEvent); virtual;
   procedure WriteResult;
   procedure NextResult (SearchError: Boolean);
   procedure Testing;
 end;

 PStatistics = ^TStatistics;
 TStatistics = object (TDialog)
   NetIn       : PInputLine;
   NetRules    : PInputLine;
   NetOut      : PInputLine;
   PatternNo   : PInputLine;
   LearnEpochs : PInputLine;
   LearnError  : PInputLine;
   Performance : PInputLine;

   Task        : PTask;
   Stats       : PStatTask;
   Net         : PNefclass;
   NewNet      : Boolean;
   constructor Init (ANet: PNefclass; ATask: PTask; AStats: PStatTask;
                     IsNewNet: Boolean);
   destructor Done; virtual;
   procedure HandleEvent (var Event: TEvent); virtual;
   procedure PlotFiles;
   procedure WriteResult;
 end;

implementation
(*========================= Methods of TTrainer ===========================*)
(*-------------------------------------------------------------------------*)
(* Graphically display the fuzzy sets of up to four variables on screen    *)
(* during the learning process. Not a method of TTrainer, because this     *)
(* will not work for some unknown reason.                                  *)
(*-------------------------------------------------------------------------*)
procedure Display_Fuzzy_Sets (Net: PNefclass; T: PTask;
                              start: integer; epoch: integer;
                              error: real; misclass: integer;
                              continous, switched, new_state,
                              single_step, halted: boolean);
 const
  ncolor = 14; (* number of colors for fuzzy sets *)
 var
  max_x, max_y, sep_y,
  stat_y, x, y,
  x1, x2, y1, y2,
  dx, dy, wx, wy        : word;
  l                     : real; (* left boundary of domain (minimum) *)
  s                     : real; (* scaling factor for width of fuzzy sets *)
  fx, fy                : integer;
  i, j, n               : integer;
  Atext                 : string[80];
  vartext               : string[10];
  TheColor              : word;   (* current drawing color *)
  BKColor               : word;
  TextColor             : word;
  HighColor             : word;
  HeaderColor           : word;
  FrameColor            : word;
  ErrorColor            : word;
  ColorTable            : array[0..ncolor-1] of integer; (* colors for classes *)

 begin
  (* We assume to be in graphics mode! *)
  (* We also assume 640x480 VGA, perhaps we will make it flexible later *)

  max_x := 639; max_y := 479; sep_y := 2 * max_y div 3 + 10;
  stat_y := max_y - 80;
  x1 := 20; x2 := 300; y1 := 50; y2 := 150;
  dx := 320; dy := 150; wx := 280; wy := 100;

  (* Setting the colors *)
  BKColor     := LightGray;
  TextColor   := Blue;
  FrameColor  := Blue;
  HeaderColor := Red;
  HighColor   := Yellow;
  ErrorColor  := Red;

  (* Initialize ColorTable for class colors *)
  TheColor := FrameColor;
  for i := 0 to ncolor-1 do
   begin
    TheColor := (TheColor + 1) mod 16;
    while TheColor in [BKColor, FrameColor] do (* don't use FrameColor or background *)
     TheColor := (TheColor + 1) mod 16;
    ColorTable[i] := TheColor;
   end;


  if continous or switched or new_state
    then begin
     if switched
       then begin
        (* draw frame *)
        SetBKColor (BKColor); (* background color *)
        ClearViewPort;
        SetColor (FrameColor);        (* frame color *)
        SetLineStyle(SolidLn,0,ThickWidth);
        Rectangle (0,0,max_x, max_y);
        Line (0, y1-15, max_x, y1-15);

        (* write header *)
        SetTextJustify (CenterText, BottomText);
        SetTextStyle (TriplexFont,HorizDir,2);
        MoveTo (max_x div 2 , (y1-25));
        SetColor(HeaderColor);  (* text color for header *)
        OutText ('NEFCLASS-PC');
        SetTextJustify (RightText, BottomText);
        SetTextStyle (SmallFont,HorizDir,5);
        MoveTo (max_x - 5 , (y1-25));
        OutText ('(c) TU Braunschweig, 1994');
        SetTextJustify (LeftText, BottomText);
        SetTextStyle (SmallFont,HorizDir,5);
        MoveTo (5 , (y1-25));
        OutText ('Version ' + VersionString + ', ' + DateString);

        (* write menu *)
        SetTextStyle (SmallFont,HorizDir,4);
        SetTextJustify (LeftText, BottomText);
        SetColor(TextColor); (* text color for menu *)
        y := sep_y;     (* separator graphics - status *)
        Line (0, y, max_x, y);
        y := stat_y; (* separator status - menu *)
        Line (0, y, max_x, y);
        y := stat_y + 10; (* begin for menu text *)
        for i := 1 to 4
         do begin
          MoveTo (64, y + 15 * i);
          OutText (DisplayMenu[i]);
         end;
       end (* if switched then *)
       else begin
        (* we have already been in graphics mode, just delete graphs/status*)
        SetViewPort(2,y1-12,max_x-2,sep_y-2,ClipOn);
        ClearViewPort;
        SetViewPort(2,sep_y+2,max_x-2,stat_y-2,ClipOn);
        ClearViewPort;
        SetViewPort (0,0,max_x,max_y,ClipOn);
       end;

     (* write the status *)
     SetTextStyle (SmallFont,HorizDir,4);
     SetColor(HighColor); (* text color for status *)
     str (Epoch: 6, VarText);
     AText := 'Epoch: ' + VarText;
     str (error:10:6, VarText);
     AText := AText + ', Error: ' + VarText;
     str ((error/ T^.count):10:6, VarText);
     AText := AText + ', mean error: ' + VarText;
     OutTextXY (120, sep_y + 20, AText);

     str (misclass, VarText);
     AText := VarText + ' patterns out of ';
     str (T^.count, VarText);
     AText := AText + VarText + ' were misclassified';
     OutTextXY (120, sep_y + 40, AText);

     SetColor (HeaderColor); (* text color for display status *)
     if continous
       then AText := 'updating display    *   '
       else AText := 'display is frozen   *   ';
     if single_step
       then AText := AText + 'single step mode   *   '
       else AText := AText + 'continous mode     *   ';
     if halted
       then AText := AText + 'learning is paused'
       else AText := AText + 'system is running';
     OutTextXY (120, sep_y + 60, AText);

     (* now draw the graphs *)
     if Net^.L1 > 4
       then n := 4
       else n := Net^.L1;

     for i :=  1 to n do begin
      if i mod 2 = 0
        then fx := 1
        else fx := 0;
      if i > 2
        then fy := 1
        else fy := 0;

      SetColor (FrameColor);
      Line (x1+fx*dx, y1+fy*dy, x1+fx*dx, y2+fy*dy);
      Line (x1+fx*dx, y2+fy*dy, x2+fx*dx, y2+fy*dy);
      OutTextXY (x1+fx*dx, y2+15+fy*dy, Net^.AntecBase[i + start - 1].VName);
      with Net^.AntecBase[i + start - 1] do begin
       l := T^.min_input[i + start - 1];
       s := wx / (T^.max_input[i + start - 1] - l);
       for j := 1 to count do
        with FSets[j] do begin
         SetColor (ColorTable[(start+i-2) mod ncolor]);
         if left_shouldered
           then y := y1+fy*dy  (* dms = 1 (Degree of MemberShip) *)
           else y := y2+fy*dy; (* dms = 0 *)
         MoveTo (x1+fx*dx + round ((a - l) * s), y);
         LineTo (x1+fx*dx + round ((b - l) * s), y1+fy*dy); (* dms = 1 *)
         if right_shouldered
           then y := y1+fy*dy  (* dms = 1 *)
           else y := y2+fy*dy; (* dms = 0 *)
         LineTo (x1+fx*dx + round ((c - l) * s), y);
        end; (* with FSets *)
      end; (* with Net^.AntecBase *)
     end; (* for i *)
    end; (* if continous or switched or new_state then *)
 end; (* Display_Fuzzy_Sets *)


(*-------------------------------------------------------------------------*)
(*                  Constructor Init of the Trainer                        *)
(*-------------------------------------------------------------------------*)
constructor TTrainer.Init (ANet: PNefclass; ATask : PTask; LP: PRLearnParam);
 var
  R         : TRect;
  HScrollbar: PScrollbar;
  VScrollbar: PScrollbar;
 begin
  R.Assign (0,0,60,17);
  inherited Init (R, 'Learning');
  Options := Options or ofCentered;
  HelpCtx := $F000;
  R.Assign (2,2,58,11);
  TextWin := New(PWindow, Init(R, 'Status', wnNoNumber));
  TextWin^.Flags := 0;
  TextWin^.SetState(sfShadow, false);
  HScrollbar := TextWin^.StandardScrollBar(sbHorizontal or sbHandleKeyboard);
  TextWin^.Insert (HScrollbar);
  VScrollbar := TextWin^.StandardScrollBar(sbVertical or sbHandleKeyboard);
  TextWin^.Insert (VScrollbar);
  R.Assign(1,1,55,8);
  Terminal := New(PTerminal, Init(R, HScrollbar, VScrollbar, 8192));

  R.Assign( 3, 12, 14, 14);
  Insert(New(PButton, Init(R, '~U~pdate', cmLearnUpdate, bfDefault)));
  R.Assign(17, 12, 28, 14);
  Insert(New(PButton, Init(R, '~S~tep', cmLearnStep, bfNormal)));
  R.Assign(31, 12, 42, 14);
  Insert(New(PButton, Init(R, '~P~ause', cmLearnHalt, bfNormal)));
  R.Assign(45, 12, 56, 14);
  Insert(New(PButton, Init(R, '~R~esume', cmLearnResume, bfNormal)));

  R.Assign( 3, 14, 15, 16);
  Insert(New(PButton, Init(R, '~C~lose', cmLearnClose, bfNormal)));
  R.Assign(17, 14, 28, 16);
  Insert(New(PButton, Init(R, '~F~req.', cmLearnFreq, bfNormal)));
  R.Assign(31, 14, 42, 16);
  Insert(New(PButton, Init(R, '~G~raphics', cmLearnGraphics, bfNormal)));
  R.Assign(45, 14, 56, 16);
  Insert(New(PButton, Init(R, '~H~elp', cmHelp, bfNormal)));

  If Application^.ValidView(Terminal) <> nil
    then begin
     AssignDevice (Term, Terminal);
     Rewrite (Term);
     TextWin^.Insert (Terminal);
     Insert (TextWin);
    end;
  Net         := ANet;
  LearnTask   := ATask;
  from        := 1;
  NofInputs   := 0;
  LParam      := LP;
  switched    := false;
  waiting     := false;
  continous   := false;
  new_state   := false;
  halted      := false;
  ForcedStop  := false;
  LearnEnd    := false;
  ErrMsgClose := false;
 end; (* TTrainer.Init *)

destructor TTrainer.Done;
 begin
  inherited done;
 end; (* TTrainer.Done *)

procedure TTrainer.HandleEvent (var Event: TEvent);
 begin
  (* when in graphics mode, we must catch PgUp and PgDown first *)
  if not LParam^.TextDisplay and ((Event.What and evKeyDown) <> 0)
    then begin
     case Event.KeyCode of
      kbPgUp : if (from <= NofInputs - 4)
                 then begin
                  inc (from);
                  waiting := false;
                  update  := true;
                  ClearEvent (Event);
                 end;
      kbPgDn : if (from > 1)
                 then begin
                  dec (from);
                  waiting := false;
                  update  := true;
                  ClearEvent (Event);
                 end;
     end; (* case *)
    end; (* then *)

  inherited HandleEvent(Event);

  if Event.What = evCommand
   then begin
    case Event.Command of
      cmLearnHalt    : begin
                        update := not halted;
                        new_state := update;
                        halted := true;
                       end;
      cmLearnClose   : if halted and LParam^.TextDisplay
                         then ForcedStop := True
                         else if LearnEnd and LParam^.TextDisplay
                                then Close
                                else ErrMsgClose := True;
      cmLearnGraphics: if LParam^.TextDisplay and not LearnEnd
                         then SwitchToGraphics
                         else if not LParam^.TextDisplay
                                then SwitchToText;
      cmLearnStep    : begin
                        LParam^.single_step := not LParam^.single_step;
                        waiting := false;
                        update := true;
                        new_state := true;
                       end;
      cmLearnUpdate  : begin
                        waiting := false;
                        update  := true;
                        new_state := true;
                       end;
      cmLearnResume  : begin
                        update := true;
                        new_state := true;
                        halted := false;
                       end;
      cmLearnFreq    : if LParam^.TextDisplay
                         then (* FrequencyDialog *)
                         else begin
                          continous := not continous;
                          new_state := true;
                          update := true;
                         end;
    end; (* case *)
    ClearEvent(Event);
   end;
 end; (* TTrainer.HandleEvent *)

procedure TTrainer.SwitchToGraphics;
 var
  grdriver, grmode  : integer;
 begin
  LParam^.TextDisplay := False;
  grdriver := detect;
  InitGraph(grdriver, grmode, PathToBGI);
  ClearViewPort;
{  GraphicsOn ('surface.plt', PathToBGI);}
  switched := true;
  update := true;
  DisableCommands ([cmHelp]);
 end; (* TTrainer.SwitchToGraphics *)

procedure TTrainer.SwitchToText;
 begin
  if Not LParam^.TextDisplay
    then begin
     CloseGraph;
{     GraphicsOff;}
     Application^.SetScreenMode(smCO80);
     Application^.ReDraw;
     LParam^.TextDisplay := true;
     switched := true;
     update := true;
     EnableCommands([cmHelp]);
    end;
 end;

(*-------------------------------------------------------------------------*)
(* This method displays the statistics of the current epoch in text mode   *)
(*-------------------------------------------------------------------------*)

procedure  TTrainer.Display_Text_Update (Epoch: integer; Error: real;
                                         misclass, count: integer);
 begin
  writeln (Term, 'Epoch: ', Epoch, ', Error: ', Error: 10:6,
           ', Mean Error: ', Error / count:10:6);
  writeln (Term, misclass, ' patterns out of ', count,
           ' were misclassified.');
  writeln (Term, '-----------------------------------------------------');
  if halted
    then begin
      writeln (Term, 'Learning is halted!');
      writeln (Term, 'Choose RESUME to go on, or CLOSE to quit.');
      writeln (Term,'-----------------------------------------------------');
    end;
 end; (* Display_Text_Update *)

(*-------------------------------------------------------------------------*)
(* Train the fuzzy perceptron with the given learning task                 *)
(*-------------------------------------------------------------------------*)
procedure TTrainer.Training (NewNet: Boolean);

 var
  Event             : TEvent;
  PatternError      : real;
  Error             : real;
  MinError          : real;
  MinErrEpoch       : LongInt;
  ErrorFile         : text;
  EIndex            : integer;
  Epoch             : LongInt;
  EndEpoch          : LongInt;
  i                 : LongInt;
  Errorbuffer       : TBuffer;
  created, wanted   : integer;
  prior             : integer;
  misclass          : integer;
  StopCriterion     : boolean;
  NewErrorFile      : boolean;

 (* Local procedure to train one pattern from the collection *)
 procedure Cycle (APattern: PPattern); far;
  var
   Class : integer;
   Maximum: real;
   ambiguous : boolean;
   correct : boolean;
   i       : integer;

  begin
   Net^.Propagate (APattern, LearnTask^.min_input);
   PatternError := Net^.FEBP (APattern, LearnTask^.min_input,
                              LearnTask^.max_input, LParam);
   Error := Error + PatternError;
  end; (* Cycle*)

 begin
  Net^.Saved := False;
  NofInputs := Net^.L1;
  writeln (Term, 'Training the NEFCLASS network with ',
           LearnTask^.count, ' patterns.');

  Epoch := 0; EIndex := 0;  MinError := 1.0E10; MinErrEpoch := 0;
  StopCriterion := false; EndEpoch := Epoch + Lparam^.EndEpoch;
  LearnTask^.Tested := False;

  NewErrorFile := Net^.Epochs = 0;
  OpenErrorFile (ErrorFile, NewErrorFile);
  if NewErrorFile
    then WriteFileHeader (ErrorFile,
                          'Protocol for rule creation and training');


  if NewNet or Net^.LearnRules or LParam^.ReLearnRules
    then begin
     if NewNet then Net^.InitNet (LearnTask);
     prior := Net^.L2;
     writeln (Term);
     write (Term, 'Number of prior rules: ', prior, '. ');
     writeln (Term, 'Now creating a rule base ...');
     writeln (Term);

     (*--- create a rule base first ---*)
     created := 0; wanted := 0;
     Writeln (ErrorFile, '% There are ', prior, ' prior rules.');
     Net^.CreateRuleBase (created, wanted, LearnTask,
                          (LParam^.RuleLearning = simple));
     WriteRuleCreators (ErrorFile, RCIndex, created);
     Writeln (ErrorFile, '% An initial rule base with ', Net^.L2,
              ' rules was created.');
     Net^.WriteRuleBase (ErrorFile);
     writeln (Term, 'An initial rule base with ', Net^.L2,
              ' rules was created.');
     if wanted > 0
       then begin
        writeln (Term, 'The system tried to create ', wanted,
                       ' more rules.');
        writeln (Term, 'Some of these were possibly identical.');
       end;

     writeln (Term, 'Press RESUME to go on.');
     halted := true;
     repeat
      GetEvent (Event);
      HandleEvent (Event);
     until not halted;

     if LParam^.RuleLearning <> simple
       then begin
        writeln (Term, 'The maximum number is ', Net^.max_rules, '. ');
        write (Term, 'Evaluating rule base, selecting best rules ');
        if LParam^.RuleLearning = Best
          then writeln (Term, '...')
          else writeln (Term, 'per class ...');
        Net^.ReconnectRules (LearnTask, ErrorFile);
        Net^.EvaluateRules (LearnTask, ErrorFile, LParam^.RuleLearning);
        writeln (ErrorFile, '% The ', Net^.L2,
                 ' best rules have been selected.');
        Net^.WriteRuleBase (ErrorFile);
        writeln (ErrorFile);
        writeln (Term, 'The ', Net^.L2,
                       ' best rules have been selected.');
        writeln (Term, 'Press RESUME to go on.');
        halted := true;
        repeat
         GetEvent (Event);
         HandleEvent (Event);
        until not halted;
       end;
     writeln (Term);

    end (* if NewNet then *)
    else begin
     writeln (Term);
     writeln (Term, 'Using existing rule base.');
     writeln (Term);
    end; (* if NewNet ... else *)

    continous := true; update := true;
    (* check if user wants graphic display *)
    if not LParam^.TextDisplay then SwitchToGraphics;


    if NewErrorFile
      then begin
       writeln (ErrorFile, '% Protocol of global error and number of ',
                'misclassifications');
       writeln (ErrorFile);
      end;

    repeat (*--- training loop ---*)

     waiting := LParam^.single_step;

     If EventAvail (* Did someone ring the bell? *)
       then begin
        GetEvent (Event);
        HandleEvent (Event);
       end;

     if not (waiting or halted) (* if in single step mode or halted, *)
        or update               (* then wait for message from user *)
       then begin
        if not (waiting or halted)
          then begin
           Error := 0;

           LearnTask^.Patterns^.ForEach (@Cycle);

           Inc (Epoch);

           if LParam^.StopCheck
             then begin
              if MinError > Error
                then begin
                 MinError := Error;
                 MinErrEpoch := Epoch;
                end
                else StopCriterion := (Epoch - MinErrEpoch > LParam^.ErrEpochs);
             end;
          end; (* if not (waiting or halted) *)
        (* check if screen and error protocol must be updated *)
        if (Epoch = 1) or (Epoch mod LParam^.UpdateRate = 0) or
           (Epoch = EndEpoch) or update or StopCriterion
          then begin
           Net^.Testing (LearnTask);
           misclass := LearnTask^.MisClass;
           ErrorBuffer[EIndex, 1] := Epoch + Net^.Epochs;
           ErrorBuffer[EIndex, 2] := Error;
           ErrorBuffer[EIndex, 3] := misclass;
           inc (EIndex);
           if EIndex = maxbuf
            then begin
             WriteErrorFile (ErrorFile, ErrorBuffer, EIndex);
             EIndex := 0;
            end; (* if EIndex ... then *)
           if LParam^.TextDisplay
            then Display_Text_Update (Epoch + Net^.Epochs, Error,
                                      misclass, LearnTask^.count)
            else Display_Fuzzy_Sets (Net, LearnTask, from,
                                     Epoch + Net^.Epochs, Error, misclass,
                                     continous, switched, new_state,
                                     LParam^.single_step, halted);
           end; (* if Epoch... then *)
       end; (* if not waiting or halted *)

      update := False;
      switched := False;
      new_state := False;

    until ((Error / LearnTask^.Count) <= LParam^.StopError) or
          (Epoch >= EndEpoch) or (misclass <= LParam^.MisClass_OK) or
          ForcedStop or StopCriterion;

    (*--- end of trainings loop ---*)

    halted   := false;
    LearnEnd := true;
    Net^.Epochs := Net^.Epochs + Epoch;
    LearnTask^.Epochs := LearnTask^.Epochs + Epoch;
    LearnTask^.Error := Error;
    LearnTask^.MisClass := MisClass;

    WriteErrorFile (ErrorFile, ErrorBuffer, EIndex);
    CloseErrorFile (ErrorFile);
    if not LParam^.TextDisplay
      then SwitchToText;
    if ForcedStop
      then Close
      else begin
       writeln(Term);
       writeln(Term,'++++++++++++++++++ Final result +++++++++++++++++++++');
       Display_Text_Update (Net^.Epochs, Error,
                            misclass,LearnTask^.count);
       writeln (Term);
       if StopCriterion
         then begin
          writeln (Term, 'Training was stopped, because the smallest error');
          writeln (Term, 'was not decreased for more than ',
                          LParam^.ErrEpochs ,' epochs');
          writeln (Term, 'The smallest Error ', MinError:0:4,
                         ' occured in epoch ',
                         MinErrEpoch + Net^.Epochs - Epoch, '.');
          writeln (Term);
         end;
       writeln (Term, 'Training completed. Press CLOSE to remove window');
      end;
 end; (* Training *)

(*====================== Methods of TResultWindow =========================*)

constructor TResultWindow.Init (ARect: TRect; ARange: integer;
                                OBuf, TBuf: PLineBuffer);
 begin
  inherited Init (ARect, '' , wnNoNumber);
  Flags := 0;
  SetState(sfShadow, false);
  HScrollbar := StandardScrollBar(sbHorizontal or sbHandleKeyboard);
  HScrollbar^.SetRange (0, ARange);
  Insert (HScrollbar);
  OutBuffer := OBuf;
  TargetBuffer := TBuf;
 end;

(*-------------------------------------------------------------------------*)

destructor TResultWindow.Done;
 begin
  inherited Done;
 end;

(*-------------------------------------------------------------------------*)
procedure TResultWindow.CopyLine (Buf: PLineBuffer; var ResBuf: TLineBuffer;
                                  Start, Stop, Chunk: integer);
 var
   i, j, k : integer;
 begin
   if (Size.X-2) div chunk > HScrollbar^.Max + 1
     then i := 0
     else i := start * chunk;
   k := 0;
   while (k < stop) and (i div chunk < HScrollbar^.Max + 1) do
    begin
     for j := 0 to chunk-1 do
      ResBuf[k+j] := Buf^[i+j];
     inc (i, chunk);
     inc (k, chunk);
    end;
 end; (* TResultWindow.CopyLine *)

(*-------------------------------------------------------------------------*)

procedure TResultWindow.Draw;

 var
  i       : integer;
  class   : integer;
  maximum : real;
  PError  : real;
  Outline : TLineBuffer;
  X       : Word;

 begin
  X := (5 * (HScrollbar^.Max + 1));
  if X > (Size.X-2) then X := Size.X-2;
  inherited Draw;
  MoveChar (OutLine, ' ', GetColor(1), 5*maxunits);
  CopyLine (OutBuffer, OutLine, HScrollbar^.Value, Size.X-2, 5);
  WriteLine (1, 1, X, 1, OutLine);
  MoveChar (OutLine, ' ', GetColor(1), 5*maxunits);
  CopyLine (TargetBuffer, OutLine, HScrollbar^.Value, Size.X-2, 5);
  WriteLine (1, 2, X, 1, OutLine);
 end;

(*-------------------------------------------------------------------------*)

procedure TResultWindow.HandleEvent (var Event: TEvent);
 begin
  inherited HandleEvent (Event);
  if Event.What = evBroadcast
   then begin
    case Event.Command of
     cmScrollBarChanged : if (Size.X-2) div 5 < HScrollbar^.Max + 1
                            then begin
                             if (HScrollbar^.Max - HScrollbar^.Value) <
                                (Size.X-2) div 5
                               then HScrollbar^.SetValue (HScrollbar^.Max + 1
                                     - (Size.X-2) div 5);
                             Draw;
                            end
                            else begin
                             HScrollbar^.SetValue(0);
                            end;
    end; (* case *)
   end;
 end; (* TResultWindow.HandleEvent *)

(*========================== Methods of TTester ===========================*)

(*-------------------------------------------------------------------------*)
(*                         Constructor of TTester                          *)
(*-------------------------------------------------------------------------*)

constructor TTester.Init (ANet: PNefclass; ATask: PTask);
 var
  R         : TRect;
 begin
  R.Assign (0,0,60,16);
  inherited Init (R, 'Testing');
  Options := Options or ofCentered;
  HelpCtx := $F000;

  New (OutBuf);
  New (TargetBuf);
  Net         := ANet;
  Task        := ATask;
  patno       := 0;

  R.Assign(2, 1, 10, 2);
  Insert (New(PStaticText, Init(R,'Pattern:')));
  R.Assign(12, 1, 20, 2);
  New (PatternNumber, Init(R, 8));
  PatternNumber^.Options := PatternNumber^.Options and not ofSelectable
                            and not ofFirstClick;
  Insert (PatternNumber);

  R.Assign(31, 1, 48, 2);
  Insert (New(PStaticText, Init(R,'Number of Errors:')));
  R.Assign(50, 1, 58, 2);
  New (ErrorNumber, Init(R, 8));
  ErrorNumber^.Options := ErrorNumber^.Options and not ofSelectable
                          and not ofFirstClick;
  Insert (ErrorNumber);

  R.Assign(2, 3, 10, 4);
  Insert (New(PStaticText, Init(R,'P_Error:')));
  R.Assign(12, 3, 20, 4);
  New (PatternError, Init(R, 8));
  PatternError^.Options := PatternError^.Options and not ofSelectable
                            and not ofFirstClick;
  Insert (PatternError);

  R.Assign(37, 3, 48, 4);
  Insert (New(PStaticText, Init(R,'Mean Error:')));
  R.Assign(50, 3, 58, 4);
  New (MeanError, Init(R, 8));
  MeanError^.Options := MeanError^.Options and not ofSelectable
                          and not ofFirstClick;
  Insert (MeanError);

  R.Assign (2,5,58,9);
  New(TextWin, Init(R, Net^.L3 - 1, OutBuf, TargetBuf));
  Insert (TextWin);

  R.Assign(2, 11, 9, 12);
  Insert(New(PStaticText, Init(R, 'Result:')));

  R.Assign (9, 11, 33, 12);
  Result := New(PRadioButtons, Init(R,
  NewSItem('Correct',
  NewSItem('False', nil))));
  Result^.Options := Result^.Options and not ofSelectable
                                     and not ofFirstClick;
  Insert(Result);

  R.Assign(44, 11, 58, 13);
  Insert(New(PButton, Init(R, 'Next ~E~rror', cmTestError, bfNormal)));

  R.Assign( 1, 13, 9, 15);
  Insert(New(PButton, Init(R, '~N~ext', cmTestNext, bfDefault)));
  R.Assign(13, 13, 21, 15);
  Insert(New(PButton, Init(R, '~P~rev', cmTestPrev, bfNormal)));
  R.Assign(25, 13, 33, 15);
  Insert(New(PButton, Init(R, '~S~ave', cmTestSave, bfNormal)));
  R.Assign(37, 13, 46, 15);
  Insert(New(PButton, Init(R, '~C~lose', cmTestClose, bfNormal)));
  R.Assign(50, 13, 58, 15);
  Insert(New(PButton, Init(R, '~H~elp', cmHelp, bfNormal)));

  Testing;
  NextResult (False);
 end;

(*-------------------------------------------------------------------------*)

destructor TTester.Done;
 begin
  Dispose (OutBuf);
  Dispose (TargetBuf);
  inherited Done;
 end;

(*-------------------------------------------------------------------------*)

procedure TTester.HandleEvent (var Event: TEvent);
 begin
  inherited HandleEvent(Event);
  if Event.What = evCommand
   then begin
    case Event.Command of
      cmTestError: begin
                    patno := (patno + 1) mod Task^.Patterns^.Count;
                    NextResult (True);
                    TextWin^.Draw
                   end;
      cmTestNext : begin
                    patno := (patno + 1) mod Task^.Patterns^.Count;
                    NextResult (False);
                    TextWin^.Draw
                   end;
      cmTestPrev : begin
                    patno := (patno - 1);
                    if patno < 0 then patno := Task^.Patterns^.Count - 1;
                    NextResult (False);
                    TextWin^.Draw
                   end;
      cmTestSave : WriteResult;
      cmTestClose: Close;
    end; (* case *)
   end; (* then *)
 end;

procedure FileClose (var T: Text);
 begin
  close (T);
 end;

(*-------------------------------------------------------------------------*)
(* Write the test result into a file.                                      *)
(*-------------------------------------------------------------------------*)
procedure TTester.WriteResult;

 var
  Term      : text;
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
  h,m,s,ss  : word; (* to get the time *)
  min       : string[2];
  y,mo,d,w  : word; (* to get the date *)


const
  FDOptions: Word = fdOKButton or fdHelpButton;

 procedure Cycle (APattern: PPattern); far;
  var
   correct : Boolean;
   ambiguous: boolean;
   i       : integer;
   Class   : integer;
   Maximum : real;
   PError  : real;

  begin
   PError := 0;
   Net^.Propagate (APattern, Task^.min_input);
   correct := Net^.CheckOutput (APattern, Class, Maximum, ambiguous);
   write (Term, APattern^.Ident:4, ') output: ');
   for i := 1 to Net^.L3 do
    begin
     if ((i-1) mod 16 = 0) and (i <> 1)
       then begin
        writeln(Term);
        write (Term, ' ':14);
       end;
     write (Term, Net^.OutLayer[i]:5:2);
     if Task^.ClassInfo
       then PError := PError +
                      sqr (APattern^.OutPattern[i] - Net^.OutLayer[i]);
    end;
   writeln (Term);

   if Task^.ClassInfo
     then begin
      write (Term, 'target: ':14);
      for i := 1 to Net^.L3 do
       begin
        if ((i-1) mod 16 = 0) and (i <> 1)
          then begin
           writeln(Term);
           write (Term, ' ':14);
          end;
        write (Term, APattern^.OutPattern[i]:5:2);
       end; (* for i *)
      writeln (Term);
      write (Term, 'Class: ', class:2, ', error: ', PError:5:3);
      if correct
        then writeln (Term, ', classification is correct')
        else writeln (Term, ', classification is false');
     end (* if Task^.ClassInfo then *)
     else begin
      write (Term,  'Class: ', class:2);
      if class = 0
        then writeln (Term, ', pattern cannot be classified')
        else writeln (Term);
     end;

   writeln (Term);
  end; (* Cycle *)

 begin

  TheFile := '*.RES';
  New(FileDialog, Init(TheFile, 'Save result file as', '~F~ile name',
    FDOptions, 1));
  FileDialog^.HelpCTX := hcTestSave;
  if Net^.FileName <> '' then TheFile := Net^.FileName + '.RES';
  if Application^.ExecuteDialog(FileDialog, @TheFile) <> cmCancel
    then begin
     if Filename (Term, writing, TheFile)
       then begin
        Rewrite (Term);
        writeln (Term, 'Result file of NEFCLASS-PC, ', VersionString,
                       ' (c) TU Braunschweig, 1994, 1995');

        GetDate (y,mo,d,w); GetTime (h,m,s,ss);
        str (m:2, min);
        if m < 10 then min[1] := '0';
        writeln (Term,'This file was created on ', mo,'/',d,'/',y, ' at ',
                       h,':',min, '.');

        writeln (Term, 'Test result for ', Task^.Patterns^.Count,
                       ' patterns, ', Task^.MisClass,
                       ' misclassifications, ',
                       ' mean error = ', (Task^.error/Task^.Count):8:4);
        writeln (Term);

        Task^.Patterns^.ForEach(@Cycle);
        FileClose(Term);
        MessageBox ('Result file was written successfully', nil,
                    mfInformation or mfOKButton)
       end
       else MessageBox ('Error writing result file!', nil,
                         mfInformation or mfOKButton)
    end;
 end; (* TTester.WriteResult *)

(*-------------------------------------------------------------------------*)
(* Testing a fuzzy perceptron with patterns, and display the results       *)
(*-------------------------------------------------------------------------*)
procedure TTester.Testing;

 var
  Infodlg  : PDialog;
  R        : TRect;

 begin
  R.Assign (0,0,40,3);
  New (InfoDlg, Init(R, 'Working'));
  InfoDlg^.Options := InfoDlg^.Options or ofCentered;
  InfoDlg^.Flags   := InfoDlg^.Flags and not wfClose;
  R.Assign (2, 1, 36, 2);
  InfoDlg^.Insert (New(PStaticText,
                       Init(R, 'Testing all patterns. Please Wait.')));
  Application^.Insert (InfoDlg);
  Net^.Testing (Task);
  InfoDlg^.Close;
 end; (* Testing *)

(*-------------------------------------------------------------------------*)

procedure TTester.NextResult (SearchError: Boolean);

 var
  i       : integer;
  start   : integer;
  maximum : real;
  perror  : real;
  Class   : integer;
  S       : string[10];
  ResST   : string[5*maxunits];
  ResSO   : string[5*maxunits];
  APattern: PPattern;
  correct : boolean;
  ambiguous: boolean;

 begin
  start := patno;
  repeat
   APattern := Task^.Patterns^.At(start);
   Net^.Propagate (APattern, Task^.min_input);
   correct := Net^.CheckOutput (APattern, Class, Maximum, ambiguous);
   perror := 0;
   for i := 1 to Net^.L3 do
    begin
     if Task^.ClassInfo
       then perror := perror +
                      sqr(Net^.OutLayer[i] - APattern^.OutPattern[i]);
    end; (* for i *)
   start := (start + 1) mod Task^.Patterns^.Count;
  until not SearchError or (patno = start) or not correct;

  if SearchError
    then begin
     patno := start - 1;
     if patno < 0 then patno := 0;
    end;

  ResSO := ''; ResST := '';
  for i := 1 to Net^.L3 do
   begin
    str (Net^.OutLayer[i]:5:2, S);
    ResSO := ResSO + S;
    if Task^.ClassInfo
      then begin
       str (APattern^.OutPattern[i]:5:2, S);
       ResST := ResST + S;
      end;
   end; (* for i *)
  MoveCStr (OutBuf^, ResSO, GetColor(1));
  if not Task^.ClassInfo
    then ResST := 'No class information in data set';
  MoveCStr (TargetBuf^, ResST, GetColor(1));
  str (APattern^.Ident:6, S);
  PatternNumber^.SetData(S);
  str (Task^.MisClass:6, S);
  ErrorNumber^.SetData (S);
  str (perror:6:4, S);
  PatternError^.SetData (S);
  str ((Task^.Error/Task^.Count):6:2, S);
  MeanError^.SetData (S);
  if correct
    then i := 0
    else i := 1;
  Result^.SetData (i);
 end; (* TTester.NextResult *)

(*====================== Methods of TStatistics ===========================*)

(*-------------------------------------------------------------------------*)
(*                      Constructor of TStatistics                         *)
(*-------------------------------------------------------------------------*)

constructor TStatistics.Init (ANet: PNefclass; ATask: PTask;
                              AStats: PStatTask; IsNewNet: Boolean);
 const
  Blanks = '                                                            ';
 var
  R         : TRect;
  S         : string[10];
  AText     : string[80];
  Line      : string[80];
  L         : integer;

 begin
  R.Assign (0,0,75,16);
  inherited Init (R, 'Statistics');
  Options := Options or ofCentered;
  HelpCtx := $F000;

  Task := ATask;
  Stats := AStats;
  Net  := ANet;
  NewNet := IsNewNet;

  AText := ' General Information Ŀ';
  R.Assign(1, 1, 74, 2);
  Insert (New(PStaticText, Init(R, AText)));


  AText := ' Network: using ';
  Str (Net^.L1, S); L := Length (S);
  AText := AText + S + ' input units, ';
  Str (Net^.L2, S); L := L + Length (S);
  AText := AText + S + ' rule units, and ';
  Str (Net^.L3, S); L := L + Length (S);
  Line := Copy (Blanks, 1, 10 - L);
  AText := AText + S + ' output units,' + Line +  '';
  R.Assign(1, 2, 74, 3);
  Insert (New(PStaticText, Init(R, AText)));

  Str (Net^.Epochs, S);
  Line := Copy (Blanks, 1 , 41-Length(S));
  AText := '          trained for ' + S + ' Epochs.' + Line + '';
  R.Assign(1, 3, 74, 4);
  Insert (New(PStaticText, Init(R, AText)));

  Line := Copy (Blanks, 1, 60 - Length(Task^.TaskName));
  AText := ' Pattern: ' + Task^.TaskName + Line + ' ';
  R.Assign(1, 4, 74, 5);
  Insert (New(PStaticText, Init(R, AText)));

  Str (Task^.Count, S);
  AText := '          with ' + S + ' patterns';
  if Task^.ClassInfo
    then Line := ' including class information.'
    else Line := ' without class information.  ';
  Line := Line + Copy (Blanks, 1, 18 - Length(S));
  AText := AText + Line + '';

  R.Assign(1, 5, 74, 6);
  Insert (New(PStaticText, Init(R, AText)));

  Str (Task^.Epochs, S);
  Line := Copy (Blanks, 1 , 29-Length(S));
  AText := '          used for training since ' + S + ' Epochs.' + Line + '';
  R.Assign(1, 6, 74, 7);
  Insert (New(PStaticText, Init(R, AText)));

  If (Net^.L2 > 0) and not Task^.Tested
    then Net^.Testing(Task);
  If Net^.L2 > 0
   then Str (Task^.Error:0:2, S)
   else S := '--';
  L := Length (S);
  AText := '          Sum of errors: ' + S;
  If Net^.L2 > 0
   then Str (Task^.MisClass, S)
   else S := '--';
  L := L + Length (S);
  Line := Copy (Blanks, 1, 14 - L);
  AText := AText + ', number of misclassifications: ' + S + Line + '';
  R.Assign(1, 7, 74, 8);
  Insert (New(PStaticText, Init(R, AText)));

  AText := '';
  R.Assign(1, 8, 74, 9);
  Insert (New(PStaticText, Init(R, AText)));

  R.Assign(4, 10, 18, 12);
  Insert(New(PButton, Init(R, '~B~ar Charts', cmStatsBar, bfNormal)));
  R.Assign(21, 10, 36, 12);
  Insert(New(PButton, Init(R, '~P~rojections', cmStatsProj, bfNormal)));
  R.Assign(39, 10, 54, 12);
  Insert(New(PButton, Init(R, '~D~escriptive', cmStatsStat, bfNormal)));
  R.Assign(57, 10, 71, 12);
  Insert(New(PButton, Init(R, 'Plot ~F~iles', cmStatsPlot, bfNormal)));

  R.Assign(20, 13, 29, 15);
  Insert(New(PButton, Init(R, '~C~lose', cmStatsClose, bfDefault)));
  R.Assign(44, 13, 53, 15);
  Insert(New(PButton, Init(R, '~H~elp', cmHelp, bfNormal)));

 end; (* TStatistics.Init *)

(*-------------------------------------------------------------------------*)

destructor TStatistics.Done;
 begin
  inherited Done;
 end; (* TStatistics.Done *)

(*-------------------------------------------------------------------------*)
(* Graphically display the projections of the patterns set in two          *)
(* dimensions. Not a method of TStatistics, because this will not work for *)
(* some unknown reason.                                                    *)
(*-------------------------------------------------------------------------*)

procedure Project_Patterns (Net: PNefclass; T: PTask;
                            dd1, dd2: integer; NewNet, Plot: Boolean);
 const
  ncolor = 14; (* number of colors for variables *)
 var
  x, max_x, y, max_y    : word;   (* general coordinates of the screen *)
  sx, wx, sy, wy, dy    : word;   (* coordinates for drawing projections *)
  hx, hy, h1, h0        : word;   (* coordinates for drawing fuzzy sets *)
  tx, ty, tdy           : word;   (* coordinates for writing text *)
  d1, d2                : integer;(* indices of features to be plotted *)
  r                     : word;   (* length of half cross, radius of circle *)
  Thick                 : word;   (* Thickness of thick lines *)
  l                     : real;   (* left boundary of domain (minimum) *)
  s                     : real;   (* scaling factor for width of fuzzy sets *)
  i, j                  : integer;(* loop variables *)
  VarText               : string[30]; (* for writing text *)
  TheColor              : word;   (* current drawing color *)
  BKColor               : word;
  TextColor             : word;
  HighColor             : word;
  HeaderColor           : word;
  FrameColor            : word;
  ErrorColor            : word;
  Key                   : Char;   (* for user commands *)
  Range                 : UnitVector; (* pattern ranges *)
  MisClass              : integer; (* number of classification errors *)
  First                 : Boolean; (* true, if first projection is drawn *)
  Stop                  : Boolean; (* loop criterium *)
  ColorTable            : array[0..ncolor-1] of integer; (* colors for classes *)

 (*---------------------------------------------*)
 (* local procedure to process the patterns *)
 procedure Cycle (APattern: PPattern); far;
  var
   i      : integer; (* loop variable *)
   x,y    : word;    (* screen coordinates to draw patterns *)
   class  : integer; (* class of current pattern *)
   hit    : real;    (* hit degree of rule, not used *)
   correct: boolean; (* true, if pattern is classified correctly *)
   ambiguous: boolean;
   maximum: integer; (* *)

  begin
   ambiguous := false;

   (* calculate coordinates to draw pattern *)
   x := sx + round (wx * ( (APattern^.InPattern[d1] - T^.min_input[d1])
                          / range[d1]));
   y := sy + wy - round (wy * ( (APattern^.InPattern[d2] - T^.min_input[d2])
                               / range[d2]));
   (* find out the class of the pattern *)

   if Net^.L2 > 0
     then begin
      Net^.Propagate (APattern, T^.min_input);
      (* correct is only interpretable, if there is ClassInfo, *)
      (* but information on Class, Hit and ambiguous is valid  *)
      correct := Net^.CheckOutput (APattern, Class, Hit, ambiguous);
     end
     else begin
      if T^.ClassInfo (* find information in pattern if there are no rules *)
        then begin
         Class := 0;
         repeat
          inc (Class);
         until (Class >= T^.Nout) or (APattern^.OutPattern[Class] > 0);
        end
        else Class := 1; (* default, if there are no rules and no ClassInfo*)
     end;

   (* draw a cross with color selected for current class *)
   SetColor (ColorTable[(Class-1) mod ncolor]);
   Line (x-r,y,x+r,y);
   Line (x,y-r,x,y+r);

   (* check if pattern is classified correctly, if there are rules *)
   if ((Net^.L2 > 0) and T^.ClassInfo and not correct) or
      ((Net^.L2 > 0) and ((Hit = 0) or ambiguous) )
    then begin
     inc (MisClass);
     SetColor (Red);
     Circle (x,y,r);
    end;

  end; (* Cycle *)

 (*----------------------------------------------*)

 (* main routine *)
 begin
  (* We assume to be in graphics mode! *)
  (* We also assume 640x480 VGA, perhaps we will make it flexible later *)

  (* Setting the colors *)
  BKColor     := Black;
  TextColor   := White;
  FrameColor  := Blue;
  HeaderColor := Red;
  HighColor   := Yellow;
  ErrorColor  := Red;

  (* Initializing the Coordinates *)
  max_x := 639; max_y := 479;
  sx := 100; wx := 300;
  sy :=  50; wy := 300;
  hy :=  75; hx :=  75;
  d1 := dd1; d2 := dd2;
  tx := sx + wx + 7;
  ty := sy+10; tdy := 18;
  r  := 4;
  First := True; MisClass := 0;
  if not plot
    then Thick := ThickWidth
    else Thick := NormWidth;

  (* calculating pattern ranges *)
  for i := 1 to T^.Nin do
   range[i] := T^.max_input[i] - T^.min_input[i];

  SetBKColor (BKColor); (* background color *)
  (* Initialize ColorTable for class colors *)
  TheColor := Blue;
  for i := 0 to ncolor-1 do
   begin
    TheColor := (TheColor + 1) mod 16;
    while TheColor in [BKColor, ErrorColor] do (* don't use error or background *)
     TheColor := (TheColor + 1) mod 16;
    ColorTable[i] := TheColor;
   end;


  (* draw frame around screen *)
  ClearViewPort;
  SetColor (FrameColor);        (* frame color *)
  SetLineStyle(SolidLn,0,Thick);
  Rectangle (0,0,max_x, max_y);
  Line (0, sy-15, max_x, sy-15);

  (* write header *)
  SetTextJustify (CenterText, BottomText);
  SetTextStyle (TriplexFont,HorizDir,2);
  MoveTo (max_x div 2 , (sy-25));
  SetColor(HeaderColor);  (* text color for header *)
  OutText ('NEFCLASS-PC');
  SetTextJustify (RightText, BottomText);
  SetTextStyle (SmallFont,HorizDir,5);
  MoveTo (max_x - 5 , (sy-25));
  OutText ('(c) TU Braunschweig, 1994');
  SetTextJustify (LeftText, BottomText);
  SetTextStyle (SmallFont,HorizDir,5);
  MoveTo (5 , (sy-25));
  OutText ('Version ' + VersionString + ', ' + DateString);

  (* write information about projetion *)
  SetTextStyle (SmallFont,HorizDir,5);
  SetColor (HighColor);
  OutTextXY (tx, ty, '2-Dimensional Data Projections');

  SetTextStyle (SmallFont,HorizDir,4);
  SetColor (TextColor);
  OutTextXY (tx, ty + tdy, 'Name of Pattern Set:');
  OutTextXY (tx, ty + 2* tdy, T^.TaskName);
  Str (T^.Count, VarText);
  OutTextXY (tx, ty+3*tdy, 'Number of Patterns: '+ VarText);
  OutTextXY (tx, ty+5*tdy, 'Horizontal Axis:');
  OutTextXY (tx+115, ty+5*tdy, 'Vertical Axis:');

  SetLineStyle(SolidLn,0,NormWidth);
  if T^.ClassInfo or (Net^.L2 > 0)
   then begin
    SetColor(ErrorColor);
    Circle (tx+2, ty+7*tdy-2, 2);
    SetColor (TextColor);
    OutTextXY (tx+10, ty+7*tdy, '= Misclassified Pattern');
   end;

  (* if there are more than ncolor classes, then colors are not unique *)
  if (Net^.L3 > ncolor) and (T^.ClassInfo or (Net^.L2 > 0))
    then begin
     stop := true;
     Str (ncolor, VarText);
     OutTextXY (tx, ty+8*tdy, 'More than ' + VarText + ' classes');
     OutTextXY (tx, ty+9*tdy, 'Color coding of classes is not unique');
    end
    else stop := false;

  (* if there are less than ncolor classes, then explain color coding, *)
  (* but only if *)
  i := 1; j := 8;
  while not stop and (T^.ClassInfo or (Net^.L2 > 0)) do
   begin
    SetColor (ColorTable[i-1]);
    Line (tx,ty-2+j*tdy,tx+4,ty-2+j*tdy);
    Line (tx+2,ty-4+j*tdy,tx+2,ty+j*tdy);

    SetColor (TextColor);
    Str (i:2, VarText);
    if NewNet
      then VarText := VarText + ' (no name)'
      else VarText := VarText + ' (' + Net^.ClassNames[i] + ')';
    OutTextXY (tx+10, ty+j*tdy, '= Class ' + VarText);
    inc (i); inc (j);
    stop := i > Net^.L3;
   end; (* while *)

  (* write user instructions on bottom of screen *)
  SetTextJustify (CenterText, BottomText);
  SetColor (TextColor);
  OutTextXY (max_x div 2, max_y-10,
             'Use left/right to change horizontal axis, '+
             'and up/down to change vertical axis. Leave with ESC');

  (* loop for drawing the projection until user quits *)
  repeat

   (* delete coordinate system, and part of text area *)
   SetViewPort (5, sy-5, sx+wx+5, max_y-20, ClipOn);
   ClearViewPort;
   SetViewPort (0, 0, max_x,max_y, ClipOn);
   (* area where the axis labels are written *)
   SetViewPort (sx+wx+7, ty + 2 + 5*tdy, max_x-10, ty+2+6*tdy, ClipOn);
   ClearViewPort;
   SetViewPort (0, 0, max_x,max_y, ClipOn);

   (* write which variable is displayed on which axis *)
   SetTextStyle (SmallFont,HorizDir,4);
   SetTextJustify (LeftText, BottomText);
   SetColor (TextColor);
   If NewNet
    then begin
     Str (d1, VarText);
     VarText := 'Feature ' + VarText;
    end
    else VarText := Net^.AntecBase[d1].VName;
   OutTextXY (tx, ty + 6*tdy, VarText);

   If NewNet
    then begin
     Str (d2, VarText);
     VarText := 'Feature ' + VarText;
    end
    else VarText := Net^.AntecBase[d2].VName;
   OutTextXY (tx+115, ty+6*tdy, VarText);

   (* Draw coordinate system *)
   SetLineStyle(SolidLn,0,NormWidth);
   SetColor (FrameColor);
   Rectangle (sx, sy, sx+wx, sy+wy);
   Line (sx, sy, sx-hx, sy);
   Line (sx, sy+wy, sx-hx, sy+wy);
   Line (sx, sy+wy, sx, sy+wy+hy);
   Line (sx+wx, sy+wy, sx+wx, sy+wy+hy);

   (* draw projections of patterns *)
   T^.Patterns^.ForEach(@Cycle);

   (* if first projection, then write number of misclassifications *)
   if First
     then begin
      if NewNet
        then VarText := 'invalid'
        else Str (MisClass, VarText);
      SetTextStyle (SmallFont,HorizDir,4);
      SetTextJustify (LeftText, BottomText);
      SetColor (TextColor);
      OutTextXY (tx, ty+4*tdy, 'Number of Misclassifications: '+ VarText);
      First := False;
     end;

   (* If there is an initialized net, then we can draw fuzzy sets *)
   if not NewNet
    then begin
     (* fuzzy sets on horizontal axis *)
     h0 := sy + wy; h1:= h0 + hy;
     SetTextJustify (CenterText, TopText);
     SetTextStyle (SmallFont,HorizDir,4);

     with Net^.AntecBase[d1] do begin
      l := T^.min_input[d1];
      s := wx / (T^.max_input[d1] - l);
      for j := 1 to count do
       with FSets[j] do begin
        SetColor (ColorTable[(ncolor-j-2) mod ncolor]);
        if left_shouldered
          then y := h1   (* dms = 1 (Degree of MemberShip) *)
          else y := h0;  (* dms = 0 *)
        MoveTo (sx + round ((a - l) * s), y);
        x := sx + round ((b - l) * s); y := h1;
        LineTo (x, y); (* dms = 1 *)
        OutTextXY (x, y+5, Name);
        MoveTo (x,y);
        if right_shouldered
          then y := h1   (* dms = 1 *)
          else y := h0;  (* dms = 0 *)
        LineTo (sx + round ((c - l) * s), y);
       end; (* with FSets *)
     end; (* with Net^.AntecBase[d1] *)

     (* fuzzy sets on vertical axis *)
     h0 := sx; h1:= sx - hx;
     SetTextJustify (RightText, CenterText);
     SetTextStyle (SmallFont,VertDir,4);

     with Net^.AntecBase[d2] do begin
      l := T^.min_input[d2];
      s := wy / (T^.max_input[d2] - l);
      for j := 1 to count do
       with FSets[j] do begin
        SetColor (ColorTable[(ncolor-j-2) mod ncolor]);
        if left_shouldered
          then x := h1   (* dms = 1 (Degree of MemberShip) *)
          else x := h0;  (* dms = 0 *)
        MoveTo (x, sy + wy - round ((a - l) * s));
        x := h1; y := sy + wy - round ((b - l) * s);
        LineTo (x, y); (* dms = 1 *)
        OutTextXY (x-5, y, Name);
        MoveTo (x,y);
        if right_shouldered
          then x := h1   (* dms = 1 *)
          else x := h0;  (* dms = 0 *)
        LineTo (x, sy + wy - round ((c - l) * s));
       end; (* with FSets *)
     end; (* with Net^.AntecBase[d2] *)
    end; (* if not NewNet *)

   if not Plot
     then begin
      repeat (* wait for user command *)
       Key := ReadKey;
      until not KeyPressed;
     end
     else Key := #27;

   case Key of (* select new variables/features with the cursor keys *)
    #72: begin (* cursor up *)
          inc (d2);
          if d2 > Net^.L1 then d2 := 1;
         end;
    #80: begin (* cursor down *)
          dec (d2);
          if d2 < 1 then d2 := Net^.L1;
         end;
    #77: begin (* cursor left *)
          inc (d1);
          if d1 > Net^.L1 then d1 := 1;
         end;
    #75: begin (* cursor right *)
          dec (d1);
          if d1 < 1 then d1 := Net^.L1;
         end;
   end;

  until (Key = #27) or Plot; (* ESC quits *)

 end; (* Project_Patterns *)

(*-------------------------------------------------------------------------*)
(* Graphically display the bar charts of the pattern features.             *)
(* Not a method of TStatistics, because this will not work for some        *)
(* unknown reason.                                                         *)
(*-------------------------------------------------------------------------*)

procedure Bar_Charts (Net: PNefclass; T: PTask; Stats: PStatTask;
                      dd1: integer; NewNet, Plot: Boolean);

 const
  wx     = 570; (* width of coordinate system *)
  ncolor = 14;  (* number of colors for variables *)

 var
  x, max_x, y, max_y    : word;   (* general coordinates of the screen *)
  sep_t, sep_b          : word;
  sx, sy, wy, dy        : word;   (* coordinates for drawing projections *)
  hx, hy, h1, h0        : word;   (* coordinates for drawing fuzzy sets *)
  tx, ty, tdy           : word;   (* coordinates for writing text *)
  d1, d2                : integer;(* indices of features to be plotted *)
  Thick                 : word;   (* Thickness of thick lines *)
  l                     : real;   (* left boundary of domain (minimum) *)
  s                     : real;   (* scaling factor for width of fuzzy sets *)
  i, j                  : integer;(* loop variables *)
  VarText               : string[30]; (* for writing text *)
  ALine                 : string[120]; (* for writing text *)
  TheColor              : word;   (* current drawing color *)
  BKColor               : word;
  TextColor             : word;
  HighColor             : word;
  HeaderColor           : word;
  FrameColor            : word;
  ErrorColor            : word;
  Key                   : Char;   (* for user commands *)
  Range                 : UnitVector; (* pattern ranges *)
  max                   : integer;(* maximal frequency in bar chart *)
  ColorTable            : array[0..ncolor] of integer; (* colors for classes *)
  Chart                 : array[1..2,0..wx] of integer; (* frequency index *)

 (*---------------------------------------------*)
 (* local procedure to process the patterns *)
 procedure Cycle (APattern: PPattern); far;
  var
   i      : integer; (* loop variable *)
   x,y    : word;    (* screen coordinates to draw patterns *)
   class  : integer; (* class of current pattern *)
   hit    : real;    (* hit degree of rule, not used *)

  begin
   (* calculate coordinate of feature value *)
   x := round (wx * ( (APattern^.InPattern[d1] - T^.min_input[d1])
               / range[d1]));
   inc (Chart[1,x]);
   if Chart[1,x] > max then max := Chart[1,x];

   (* find out the class of the pattern *)
   if T^.ClassInfo
     then begin
      i := 0;
      repeat
       inc (i);
      until APattern^.OutPattern[i] > 0;
     end
     else i := 1;

   (* Select color for entry, ErrorColor if not unique *)
   if Chart [2,x] < 0
     then Chart[2,x] := ColorTable[(i-1) mod ncolor]
     else if Chart[2,x] <> ColorTable[(i-1) mod ncolor]
            then Chart[2,x] := ErrorColor;

  end; (* Cycle *)

 (*----------------------------------------------*)

 (* main routine *)
 begin
  (* We assume to be in graphics mode! *)
  (* We also assume 640x480 VGA, perhaps we will make it flexible later *)

  (* Setting the colors *)
  BKColor     := Black;
  TextColor   := White;
  FrameColor  := Blue;
  HeaderColor := Red;
  HighColor   := Yellow;
  ErrorColor  := Red;

  (* Initializing the Coordinates *)
  max_x := 639; max_y := 479;
  sx := 50;
  sy := 60; wy := 200;
  hy :=  75;
  sep_t := 35; sep_b := 360;
  tx := 10; ty := sep_b + 20; tdy := 15;
  d1 := dd1;
  if not plot
    then Thick := ThickWidth
    else Thick := NormWidth;

  (* calculating pattern ranges *)
  for i := 1 to T^.Nin do
   range[i] := T^.max_input[i] - T^.min_input[i];

  SetBKColor (BKColor); (* background color *)

  (* Initialize ColorTable for class colors *)
  TheColor := Blue;
  for i := 0 to ncolor-1 do
   begin
    TheColor := (TheColor + 1) mod 16;
    while TheColor in [BKColor, ErrorColor] do (* don't use ErrorColor or background *)
     TheColor := (TheColor + 1) mod 16;
    ColorTable[i] := TheColor;
   end;


  (* draw frame around screen *)
  ClearViewPort;
  SetColor (FrameColor);        (* frame color *)
  SetLineStyle(SolidLn,0,Thick);
  Rectangle (0,0,max_x, max_y);
  Line (0, sep_t, max_x, sep_t);
  Line (0, sep_b, max_x, sep_b);

  (* write header *)
  SetTextJustify (CenterText, BottomText);
  SetTextStyle (TriplexFont,HorizDir,2);
  MoveTo (max_x div 2 , (sep_t-10));
  SetColor(HeaderColor);  (* text color for header *)
  OutText ('NEFCLASS-PC');
  SetTextJustify (RightText, BottomText);
  SetTextStyle (SmallFont,HorizDir,5);
  MoveTo (max_x - 5 , (sep_t-10));
  OutText ('(c) TU Braunschweig, 1994');
  SetTextJustify (LeftText, BottomText);
  SetTextStyle (SmallFont,HorizDir,5);
  MoveTo (5 , (sep_t-10));
  OutText ('Version ' + VersionString + ', ' + DateString);

  (* write user instructions on bottom of screen *)
  SetTextJustify (CenterText, BottomText);
  SetTextStyle (SmallFont,HorizDir,4);
  SetColor (TextColor);
  OutTextXY (max_x div 2, max_y-10,
             'Use up/down to select a new feature. Leave with ESC');


  repeat (* draw bar charts until user quits *)

   (* initialize the frequency and color index *)
   for i := 0 to wx do
    begin
     Chart[1,i] := 0;
     Chart[2,i] := -1;
    end;

   (* delete coordinate system, and part of text area *)
   SetViewPort (5, sep_t+5, max_x-5, sep_b-5, ClipOn);
   ClearViewPort;
   SetViewPort (0, 0, max_x,max_y, ClipOn);
   SetViewPort (5, sep_b+5, max_x-5, max_y-20, ClipOn);
   ClearViewPort;
   SetViewPort (0, 0, max_x,max_y, ClipOn);

   (* Draw coordinate system *)

   SetColor (HighColor);
   SetTextStyle (SmallFont,HorizDir,5);
   SetTextJustify (CenterText, BottomText);

   Str (d1, VarText);
   if not NewNet
    then VarText := VarText + ' (' + Net^.AntecBase[d1].VName + ')';
   OutTextXY (max_x div 2, sy-5, 'Bar Chart of Input Feature ' + VarText);

   SetTextStyle (SmallFont,HorizDir,4);
   SetTextJustify (RightText, CenterText);
   SetLineStyle(SolidLn,0,NormWidth);
   SetColor (FrameColor);
   Rectangle (sx-1, sy-1, sx+wx+1, sy+wy+1);
   Line (sx-1,sy, sx-5, sy);
   Line (sx-1,sy+wy, sx-5, sy+wy); OutTextXY (sx-7,sy+wy, '0');

   (* calculate distribution of selected feature *)
   max := 0;
   T^.Patterns^.ForEach(@Cycle);

   Str (max, VarText);
   OutTextXY (sx-7,sy, VarText); (* write maximum frequency to axis *)

   (* write statistics of feature *)
   SetColor (TextColor);
   SetTextJustify (LeftText, BottomText);

   OutTextXY (tx, ty, 'Name of Pattern Set: ' + T^.TaskName);
   SetColor (HighColor);
   OutTextXY (tx, ty + tdy, 'Statistics for this Input Feature:');

   Str (Stats^.Mean[d1]:0:4, VarText);
   ALine := 'Mean: ' + VarText + '   Variance: ';
   Str (Stats^.Variance[d1]:0:4, VarText);
   ALine:= ALine + VarText + '   Range: ';
   Str (T^.min_input[d1]:0:4, VarText);
   ALine := ALine + VarText + ' to ';
   Str (T^.max_input[d1]:0:4, VarText);
   ALine := ALine + VarText;
   if Stats^.CorrOutIsValid
     then begin
      Str (Stats^.CorrOut[d1]:0:4, VarText);
      ALine := ALine + '   Correlation with Class Information: ' + VarText;
     end;
   OutTextXY (tx, ty + 2 * tdy, ALine);
   SetColor (TextColor);

   if T^.ClassInfo
     then begin
      Str (Net^.L3, VarText);
      ALine := 'Values of different classes are drawn in' +
               ' different colors (there are ' + VarText + ' classes, ' +
               ' colors are ';
      if Net^.L3 > ncolor
        then ALine := ALine + 'not unique)'
        else ALine := ALine + 'unique)';
      OutTextXY (tx, ty + 3 * tdy, ALine);
      SetColor (ErrorColor);
      OutTextXY (tx, ty + 4 * tdy,
       'This color means that a value exists in more than one class');
      SetColor (TextColor);
     end;
   for i := 1 to wx do
    begin
     if Chart[1,i] > 0
       then begin
        SetColor (Chart[2,i]);
        MoveTo (sx+i,sy+wy);
        LineTo (sx+i,sy+wy - round (wy * Chart[1,i] / Max));
       end;
    end; (* for *)

   (* If there is an initialized net, then we can draw fuzzy sets *)
   if not NewNet
    then begin
     (* fuzzy sets on horizontal axis *)
     h0 := sy + wy; h1:= h0 + hy;
     SetTextJustify (CenterText, TopText);
     SetTextStyle (SmallFont,HorizDir,4);

     with Net^.AntecBase[d1] do begin
      l := T^.min_input[d1];
      s := wx / (T^.max_input[d1] - l);
      for j := 1 to count do
       with FSets[j] do begin
        SetColor (ColorTable[(ncolor-j-2) mod ncolor]);
        if left_shouldered
          then y := h1   (* dms = 1 (Degree of MemberShip) *)
          else y := h0;  (* dms = 0 *)
        MoveTo (sx + round ((a - l) * s), y);
        x := sx + round ((b - l) * s); y := h1;
        LineTo (x, y); (* dms = 1 *)
        OutTextXY (x, y+5, Name);
        MoveTo (x,y);
        if right_shouldered
          then y := h1   (* dms = 1 *)
          else y := h0;  (* dms = 0 *)
        LineTo (sx + round ((c - l) * s), y);
       end; (* with FSets *)
     end; (* with Net^.AntecBase[d1] *)
    end; (* if not NewNet *)

   if not Plot
     then begin
      repeat (* wait for user command *)
       Key := ReadKey;
      until not KeyPressed;
     end
     else Key := #27;

   case Key of (* select new features with the cursor keys *)
    #72: begin (* cursor up *)
          inc (d1);
          if d1 > T^.Nin then d1 := 1;
         end;
    #80: begin (* cursor down *)
          dec (d1);
          if d1 < 1 then d1 := T^.Nin;
         end;
   end;

  until (Key = #27) or Plot; (* ESC quits *)

 end; (* Bar_Charts *)

(*-------------------------------------------------------------------------*)

procedure TStatistics.PlotFiles;
 var
   FileDialog: PFileDialog;
   TheFile   : FNameStr;
   TestFile  : Text;
   PlotFile  : File;
   PlotBuffer: PRPlotParam;
   PlotDialog: PPlotDialog;
   project   : boolean;
   V1, V2    : integer;
   Code      : integer;

 const
   FDOptions: Word = fdOKButton or fdHelpButton;
 begin
  PlotDialog := PPlotDialog(ResFile.Get('PLOT'));
  PlotBuffer := nil;
  if PlotBuffer = nil
    then begin
     new (PlotBuffer);
     PlotBuffer^.V1 := '1';
     PlotBuffer^.V2 := '2';
     PlotBuffer^.PlotSelect := 0;
    end;
  if Application^.ExecuteDialog (PlotDialog, PlotBuffer) <> cmCancel
    then begin
     project := (PlotBuffer^.PlotSelect = 0);
     Val (PlotBuffer^.V1, V1, Code);
     Val (PlotBuffer^.V2, V2, Code);
     if V1 > Net^.L1 then V1 := 1;
     if V2 > Net^.L1 then V2 := 1;

     TheFile := '*.PLT';
     New(FileDialog, Init(TheFile, 'Plot Graphics to File', '~F~ile name',
       FDOptions, 1));
     FileDialog^.HelpCTX := hcPlotSave;
     if Task^.FileName <> '' then TheFile := Task^.FileName + '.PLT';

     if Application^.ExecuteDialog(FileDialog, @TheFile) <> cmCancel
       then begin
        if Filename (TestFile, writing, TheFile)
          then begin
           DisableCommands ([cmHelp]);
           GraphicsOn (TheFile, PathToBGI);
           if project
             then Project_Patterns (Net, Task, V1, V2, NewNet, True)
             else begin
              if not Stats^.MeanVarIsValid
                then begin
                 Stats^.Compute_All;
                end;
              Bar_Charts (Net, Task, Stats, V1, NewNet, True);
             end;
           GraphicsOff;
           Application^.SetScreenMode(smCO80);
           Application^.ReDraw;
           EnableCommands([cmHelp]);
           MessageBox ('Graphic was plotted successfully', nil,
                       mfInformation or mfOKButton)
          end (* if Filename then *)
          else MessageBox ('Could not write plot file', nil,
                           mfError or mfOKButton);
       end; (* if ExecuteDialog (File) then *)
    end; (* if ExecuteDialog (Plot) then *)
    if PlotBuffer <> nil then Dispose (PlotBuffer);
 end; (* TStatistics.PlotFiles *)

(*-------------------------------------------------------------------------*)

procedure TStatistics.HandleEvent (var Event: TEvent);
 var
  grdriver, grmode : integer;
  Infodlg  : PDialog;
  R        : TRect;
 begin
  inherited HandleEvent (Event);
  if Event.What = evCommand
   then begin
    case Event.Command of
      cmStatsProj : begin
                     grdriver := detect;
                     InitGraph(grdriver, grmode, PathToBGI);
                     ClearViewPort;
                     DisableCommands ([cmHelp]);
                     Project_Patterns (Net,Task,1,2,NewNet, False);
                     CloseGraph;
                     Application^.SetScreenMode(smCO80);
                     Application^.ReDraw;
                     EnableCommands([cmHelp]);
                    end;
      cmStatsBar  : begin
                     R.Assign (0,0,38,3);
                     New (InfoDlg, Init(R, 'Working'));
                     InfoDlg^.Options := InfoDlg^.Options or ofCentered;
                     InfoDlg^.Flags   := InfoDlg^.Flags and not wfClose;
                     R.Assign (2, 1, 34, 2);
                     InfoDlg^.Insert (New(PStaticText, Init(R,
                            'Computing Statistics. Please Wait.')));
                     Application^.Insert (InfoDlg);
                     if not Stats^.MeanVarIsValid
                       then begin
                        Stats^.Compute_All;
                       end;
                     InfoDlg^.Close;
                     grdriver := detect;
                     InitGraph(grdriver, grmode, PathToBGI);
                     ClearViewPort;
                     DisableCommands ([cmHelp]);
                     Bar_Charts (Net,Task,Stats,1,NewNet, False);
                     CloseGraph;
                     Application^.SetScreenMode(smCO80);
                     Application^.ReDraw;
                     EnableCommands([cmHelp]);
                    end;
      cmStatsStat : WriteResult;
      cmStatsPlot : PlotFiles;
      cmStatsClose: Close;
    end; (* case *)
   end; (* then *)

 end; (* TStatistics.HandleEvent *)

(*-------------------------------------------------------------------------*)
(* Write the statistics into a file.                                       *)
(*-------------------------------------------------------------------------*)
procedure TStatistics.WriteResult;

 var
  Term      : text;
  FileDialog: PFileDialog;
  TheFile   : FNameStr;

const
  FDOptions: Word = fdOKButton or fdHelpButton;

 begin

  TheFile := '*.STA';
  New(FileDialog, Init(TheFile, 'Save statistics file as', '~F~ile name',
    FDOptions, 1));
  FileDialog^.HelpCTX := hcStatSave;
  if Task^.FileName <> '' then TheFile := Task^.FileName + '.STA';

  if Application^.ExecuteDialog(FileDialog, @TheFile) <> cmCancel
    then begin
     if Filename (Term, writing, TheFile)
       then begin
        Stats^.WriteStats (Term);
        FileClose(Term);
        MessageBox ('Statistics file was written successfully', nil,
                    mfInformation or mfOKButton)
       end
       else MessageBox ('Error writing statistics file!', nil,
                         mfInformation or mfOKButton)
    end;
 end; (* TStatistics.WriteStats *)

end.