(*-------------------------------------------------------------------------*)
(*                             NEFCLASS-PC                                 *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Crisp classification by a fuzzy perceptron with fuzzy antecedents and   *)
(* crisp conclusion weights                                                *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Filename: NFCL_TSK.PAS, unit for pattern handling                       *)
(* Programmer: Ulrike Nauck                                                *)
(* Last changes: 11/02/95                                                  *)
(*-------------------------------------------------------------------------*)
(* (c) Technical University of Braunschweig, 1994                          *)
(*-------------------------------------------------------------------------*)

Unit NFCL_Tsk;

interface

uses Dos, Drivers, App, Objects, Dialogs, TextView,
     Views, MsgBox, Memory,
     NFCL_Con, NFCL_Typ;

type

  (* A single pattern has an input and an output vector (target) *)
  PPattern         = ^TPattern;
  TPattern         = object (TObject)
                     Ident     : LongInt;
                     InPattern : FeatureVector;
                     OutPattern: FeatureVector;
                     constructor Init (AIdent: LongInt;
                                       InPats, OutPats: FeatureVector);
                    end;

  (* A learning task consists of a number of patterns *)
  PTask = ^TTask;
  TTask = object (TObject)
            Patterns : PCollection;
            max_input: UnitVector; (* max and min values for each *)
            min_input: UnitVector; (* input feature *)
            Count    : integer;    (* number of patterns *)
            Nin, Nout: integer;    (* input and output dimensions *)
            ClassInfo: boolean;    (* true, if output patterns present *)
            TaskName : string[60]; (* Name of pattern set, default is filename *)
            FileName : NameStr;    (* Name of pattern file *)
            Epochs   : LongInt;    (* Number of epochs patterns were used *)
            Error    : real;       (* Sum of errors for this pattern set *)
            MisClass : integer;    (* Number of misclassifications *)
            Tested   : boolean;    (* True if pattern set tested a network *)
            constructor Init;
            destructor Done; virtual;
            function ReadPattern (var PatternFile: text;
                                  InCount, OutCount: integer): boolean;
          end;

  CorrArray = array[1..maxunits] of UnitVector;

  PStatTask = ^TStatTask;
  TStatTask = object (TObject)
                Mean: UnitVector; (* means of the features *)
                Variance : UnitVector; (* variance of the features *)
                CorrOut : UnitVector; (* correlation features - class *)
                Correlation : CorrArray; (* correlation of features *)
                Ranges : RRanges;
                Task   : PTask; (* Pattern set to compute statistics *)
                Categories  : integer; (* number of categories (features) *)
                N      : integer; (* number of cases (patterns) *)
                AllIsValid: Boolean; (* All Statistics have been computed *)
                MeanVarIsValid: Boolean; (* Means have been computed *)
                CorrOutIsValid: Boolean; (* Correlations with class have been computed *)
                CorrsIsValid: Boolean; (* Intercorreleations have been computed *)
                constructor Init (ATask: PTask);
                destructor Done; virtual;
                procedure Compute_Mean_and_Variance;
                procedure Compute_CorrOut;
                procedure Compute_Corrs;
                procedure Compute_All;
                procedure WriteStats (Var Term: Text);
              end;

implementation
(*=========================== Methods of TPattern =========================*)

constructor TPattern.Init (AIdent: LongInt;
                           InPats, OutPats: FeatureVector);
 begin
  inherited Init;
  Ident      := AIdent;
  InPattern  := InPats;
  OutPattern := OutPats;
 end;

(*============================ Methods of TTask ===========================*)

constructor TTask.Init;
 begin
  inherited Init;
  Count    := 0;
  Patterns := nil;
  Nin      := 0;
  Nout     := 0;
  ClassInfo:= True;
  TaskName := '';
  Epochs   := 0;
  Error    := 0;
  MisClass := 0;
  Tested   := False;
 end;

destructor TTask.Done;
 begin
  if Patterns <> nil then Dispose (Patterns, Done);
  inherited Done;
 end;

(*-------------------------------------------------------------------------*)
(* This methods reads a pattern file for learning or testing               *)
(* Format: [<comment>] <data area> ... END (see also procedure ReadNet)    *)
(*         <data area>: <name> | <unknown> | <inranges> | <pattern>        *)
(*         <name>     : NAME <newline> <string of max. 60 chars.>          *)
(*         <unknown>  : UNKNOWN <newline>                                  *)
(*         <inranges> : INRANGES <newline>                                 *)
(*                      <min max> ...  <newline>                           *)
(*         <pattern>  : PATTERNS <newline>                                 *)
(*                      <n. of patterns> <n. of inputs> <n. of outputs>    *)
(*                      <newline>                                          *)
(*                      <in feature> ... <out feature> ... <newline>       *)
(*-------------------------------------------------------------------------*)

function TTask.ReadPattern (var PatternFile: text;
                            InCount, OutCount: integer): boolean;
 var
  i, j, n : integer;
  Result  : boolean;
  TheEnd  : boolean;
  command : string;
  ident   : boolean;
  APat    : PPattern;
  NoInRanges: Boolean;
  InRangesCorrected: Boolean;
  WrongInRanges : Boolean;

 (*--- internal procedure to read the patterns ---*)
 procedure read_patterns (ident: boolean);
  var
   i,j     : integer;
   InPats  : FeatureVector;
   OutPats : FeatureVector;
   Id      : LongInt;
   First   : Boolean;

  begin
   First := True;
   readln (PatternFile, Count, Nin, Nout);
   if (Nin <> InCount) or
      ((Nout <> OutCount) and ClassInfo)
     then Result := False
     else begin
      Result := True;
      If not ClassInfo then Nout := 0;

      (* Delete old patterns, and create a new collection *)
      if Patterns <> nil then Patterns^.Done;
      If MaxAvail < 1.0 * SizeOf(TPattern) * Count
        then begin
         Result := False;
         MessageBox (#3'There is not enough memory'#13#3+
                     'to store all patterns.', nil,
                     mfError or mfOKButton);
        end
        else begin
         new (Patterns, Init(Count, Count Div 10 + 1));
         for i := 1 to Count do
          begin
           if ident
             then read (PatternFile, Id)
             else Id := i;
           for j := 1 to Nin do
            begin
             read (PatternFile, InPats[j]);
             if NoInRanges and First
               then begin
                min_input[j] := InPats[j];
                max_input[j] := InPats[j];
               end
               else begin
                if min_input[j] > InPats[j]
                  then begin
                   min_input[j] := InPats[j];
                   InRangesCorrected := True;
                  end;
                if max_input[j] < InPats[j]
                  then begin
                   max_input[j] := InPats[j];
                   InRangesCorrected := True;
                  end;
               end; (* else *)
            end; (* for j := 1 to Nin *)
           for j := 1 to Nout do
            read (PatternFile, OutPats[j]);
           if not ClassInfo
             then for j := 1 to OutCount do
                    OutPats[j] := 0;
           readln (PatternFile);
           new (APat, Init(Id, InPats, OutPats));
           Patterns^.Insert (APat);
           First := False;
          end; (* for i := 1 to Count *)
        end; (* if MaxAvail *)
     end; (* else *)
  end; (* read_patterns *)

 (*--- internal procedure to read ranges of input features ---*)
 procedure read_inranges;
  var
   i, j : integer;
   min, max: real;
  begin
   for i := 1 to InCount do
    begin
     read (PatternFile, min, max);
     if InRangesCorrected (* i.e. we have got ranges from the patterns *)
       then begin
        if min <= min_input[i]
          then min_input[i] := min
          else WrongInRanges := True;
        if max >= max_input[i]
          then max_input[i] := max
          else WrongInRanges := True;
       end
       else begin  (* i.e. there are no ranges yet *)
        min_input[i] := min;
        max_input[i] := max;
       end;
    end; (* for *)
   readln (PatternFile);
   NoInRanges := False;
  end; (* read_param *)

 (*-- internal function error to write error message and to return false --*)
 function FileError (command:string): boolean;
  var
   TheCommand: PString;
  begin
   TheCommand := @command;
   MessageBox (#3'Unexpected error in pattern file'#13#3+
               'while processing'#13#3'"%s".', @TheCommand,
               mfError or mfOKButton);
   FileError := false;
  end;

 (*--- internal procedure to read the task name ---*)
 procedure read_name;
  begin
   readln (PatternFile, TaskName);
  end; (* read_name *)

 (*--- body of ReadPattern ---*)
 begin

  Reset (PatternFile);
  Result := True;
  TheEnd := False;
  ident  := False;
  NoInRanges := True;
  InRangesCorrected := False;
  WrongInRanges := False;

  repeat
   command := skip_comments (PatternFile);
   if pos ('IN', command) = 1
     then begin
      read_inranges;
      if WrongInRanges
        then MessageBox (#3'The specified ranges do not encompass'#13#3+
                         'all feature values!'#13#3+
                         'That has been corrected.', nil,
                         mfError or mfOKButton);
     end
   else if pos ('ID', command) = 1
     then ident := true
   else if pos ('NO', command) = 1
     then ident := false
   else if pos ('P', command) = 1
     then begin
      read_patterns (ident);
      if not NoInranges and InRangesCorrected
        then MessageBox (#3'The specified ranges do not encompass'#13#3+
                         'all feature values!'#13#3+
                         'That has been corrected.', nil,
                         mfError or mfOKButton);
     end
   else if pos ('END',command) = 1
     then TheEnd := True
   else if pos ('UN', command) = 1
     then ClassInfo := False
   else if pos ('NA', command) = 1
     then read_name
   else Result := FileError (command);
  until TheEnd or (Result = False);

  Close (PatternFile);
  ReadPattern := Result;
 end; (* TTask.ReadPattern *)

(*========================= Methods of TStatTask ==========================*)

(*-------------------------------------------------------------------------*)

constructor TStatTask.Init (ATask: PTask);
 begin
  inherited Init;
  Task := ATask;
  AllIsValid := False;
  MeanVarIsValid := False;
  CorrOutIsValid := False;
  CorrsIsValid := False;
  N := Task^.Count;
  Categories := Task^.Nin;
  Compute_all;
 end; (* TStatTask.Init *)

(*-------------------------------------------------------------------------*)

destructor TStatTask.Done;
 begin
  inherited Done;
 end; (* TStatTask.Done *)

(*-------------------------------------------------------------------------*)

procedure TStatTask.Compute_Mean_and_Variance;
 var
  i: integer;
 procedure Cycle (APattern: PPattern); far;
  var
   i : integer;
   x : real;
  begin
   for i := 1 to Categories do
    begin
     x := APattern^.InPattern[i];
     Mean[i] := Mean[i] + x;
     Variance[i] := Variance[i] + Sqr (x);
     if x < Ranges.Min[i]
       then Ranges.Min[i] := x;
     if x > Ranges.Max[i]
       then Ranges.Max[i] := x;
    end;
  end;
 begin
  for i := 1 to Categories do
    begin
     Mean[i] := 0;
     Variance[i] := 0;
     Ranges.Min[i] := 1E38; Ranges.Max[i] := -1E38;
    end;
  Task^.Patterns^.ForEach (@Cycle);
  for i := 1 to Categories do
    begin
     Variance[i] := (Variance[i] - sqr (Mean[i]) / N) / N;
     Mean[i] := Mean[i] / N;
    end;
  MeanVarIsValid := True;
 end; (* TStatTask.Compute_Mean_and_Variance *)

(*-------------------------------------------------------------------------*)

procedure TStatTask.Compute_CorrOut;
 var
  CMean, CVar: real;
  i: integer;

 procedure Cycle2 (APattern: PPattern); far;
  var
   Class, i : integer;
  begin
    Class := 0;
    repeat
     inc (Class);
    until (Class >= Task^.Nout) or (APattern^.OutPattern[Class] <> 0);
    CMean := CMean + Class; CVar := CVar + sqr (Class);
    for i := 1 to Categories do
     CorrOut[i] := CorrOut[i] +
                   (APattern^.InPattern[i] * Class)
  end ; (* Cycle2 *)

 begin

  if Task^.ClassInfo
    then begin
     if not MeanVarIsValid then Compute_Mean_and_Variance;

     for i := 1 to Categories do
      CorrOut[i] := 0;

     CMean := 0; CVar := 0;

     Task^.Patterns^.ForEach (@Cycle2);

     CVar := (CVar - sqr (CMean) / N) / N;
     CMean:= CMean / N;

     for i := 1 to Categories do
      CorrOut[i] := (CorrOut[i] - (Mean[i] * CMean) * N ) /
                    (N * Sqrt (Variance[i]) * Sqrt (CVar));
     CorrOutIsValid := True;
    end;

 end; (* TStatTask.Compute_CorrOut *)

(*-------------------------------------------------------------------------*)

procedure TStatTask.Compute_Corrs;
 procedure Cycle (APattern: PPattern); far;
  var
   i, j : integer;
  begin
    for i := 1 to Categories do
     for j := i to Categories do
      Correlation[i,j] := Correlation[i,j] +
                          (APattern^.InPattern[i] * APattern^.InPattern[j]);
  end ; (* Cycle *)
 var
  i,j: integer;
 begin
  if not MeanVarIsValid then Compute_Mean_and_Variance;
  for i := 1 to Categories do
   for j := 1 to Categories do
    Correlation[i,j] := 0;

  Task^.Patterns^.ForEach (@Cycle);

  for i := 1 to Categories do
   for j := i to Categories do
   Correlation[i,j] := (Correlation[i,j] - (Mean[i] * Mean[j]) * N ) /
                 (N * Sqrt (Variance[i]) * Sqrt (Variance[j]));
  CorrsIsValid := True;
 end; (* TStatTask.Compute_Corrs *)

(*-------------------------------------------------------------------------*)

procedure TStatTask.Compute_All;
 begin
  if not MeanVarIsValid then Compute_Mean_and_Variance;
  if not CorrOutIsValid then Compute_CorrOut;
  if not CorrsIsValid then Compute_Corrs;
  AllIsValid := MeanVarIsValid and CorrOutIsValid and CorrsIsValid;
 end; (* TStatTask.Compute_All *)

(*-------------------------------------------------------------------------*)

procedure TStatTask.WriteStats (Var Term: Text);
 var
  Line: String[79];
  h,m,s,ss  : word; (* to get the time *)
  min       : string[2];
  y,mo,d,w  : word; (* to get the date *)
  i, j : integer;
 begin
  Line := ''+
          '';
  if not AllIsValid then Compute_All;
  Rewrite (Term);
  writeln (Term, 'Statistics file of NEFCLASS-PC, ', VersionString,
                 ' (c) TU Braunschweig, 1994, 1995');

  GetDate (y,mo,d,w); GetTime (h,m,s,ss);
  str (m:2, min);
  if m < 10 then min[1] := '0';
  writeln (Term,'This file was created on ', mo,'/',d,'/',y, ' at ',
                 h,':',min, '.');

  writeln (Term);
  writeln (Term, Line);
  writeln (Term);
  writeln (Term, 'DESCRIPTIVES OF PATTERN SET');
  writeln (Term);
  writeln (Term, 'Name: ', Task^.TaskName);
  writeln (Term, 'Number of patterns: ', Task^.Count);
  writeln (Term, 'Number of features: ', Task^.Nin);
  if Task^.ClassInfo
    then writeln (Term, 'Number of classes : ', Task^.Nout)
    else writeln (Term, 'There is no class information');
  writeln (Term, Line);
  writeln (Term, 'DESCRIPTIVES OF FEATURES');
  writeln (Term);
  for i := 1 to Categories do
   begin
    writeln (Term, 'Feature ', i:2);
    writeln (Term, '');
    writeln (Term, 'Range specified in pattern file: [',
                    Task^.min_input[i]:0:2, ', ',
                    Task^.max_input[i]:0:2, '], ');
    writeln (Term, 'Real range as found in patterns: [',
                    Ranges.min[i]:0:2, ', ', Ranges.max[i]:0:2, '].');
    writeln (Term, 'Mean: ', Mean[i]:0:2, ', Variance: ', Variance[i]:0:2, ', ',
                   'Standard Deviation: ', Sqrt (Variance[i]):0:2);
   writeln (Term);
  end; (* for i *)

  writeln (Term, Line);
  writeln (Term, 'CORRELATIONS');
  writeln (Term);
  writeln (Term, 'Inbetween features');
  writeln (Term, '');
  writeln (Term);

  write (Term, '     1  ');
  for i := 2 to Categories do
   write (Term, i:4, '  ');
  writeln (Term);

  write (Term, '');
  for i := 2 to Categories do
   write (Term, '');
  writeln (Term);

  for i := 1 to Categories do
   begin
    write (Term, i:2, '');
    for j := 1 to Categories do
     begin
       if j < i
         then write (Term, ' ':6)
         else write (Term, Correlation[i,j]:6:2);
     end; (* for j *)
    writeln (Term);
   end; (* for i *)

  writeln (Term);
  writeln (Term, 'Between features and class');
  writeln (Term, '');
  writeln (Term);
  write (Term, '     1  ');
  for i := 2 to Categories do
   write (Term, i:4, '  ');
  writeln (Term);

  write (Term, '');
  for i := 2 to Categories do
   write (Term, '');
  writeln (Term);

  write (Term, 'Cl');
  for j := 1 to Categories do
   begin
    write (Term, CorrOut[j]:6:2);
   end; (* for j *)
  writeln (Term);

 end; (* TStatTask.WriteStats *)

end.