(*-------------------------------------------------------------------------*)
(*                             NEFCLASS-PC                                 *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Crisp classification by a fuzzy perceptron with fuzzy antecedents and   *)
(* crisp conclusion weights                                                *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Filename: NFCL_TYP.PAS, unit for data types and service routines        *)
(* Programmer: Ulrike Nauck                                                *)
(* Last changes: 10/01/95                                                  *)
(*-------------------------------------------------------------------------*)
(* (c) Technical University of Braunschweig, 1994                          *)
(*-------------------------------------------------------------------------*)

Unit NFCL_Typ;

interface

uses Dos, Objects, MSGBox, NFCL_Con;

type
  FSNCStr = string[50]; (* String for fuzzy set name collection *)
  FSNStr  = string[8];  (* String for a fuzzy set name *)
  VNStr   = string[8];  (* String for a name of a variable *)
  PVNStr  = ^VNStr;
  AntStr  = string[24]; (* String to describe an antecedent *)
  PAntStr = ^AntStr;
  LabelStr = string[80]; (* String for network description *)

const

 FNames : array[0..7] of FSNCStr =
  ('not_used',                                            (* unused var.  *)
   'dontcare',                                            (* 1 fuzzy set  *)
   'small/large',                                         (* 2 fuzzy sets *)
   'small/medium/large',                                  (* 3 fuzzy sets *)
   'v_small/small/large/v_large',                         (* 4 fuzzy sets *)
   'v_small/small/medium/large/v_large',                  (* 5 fuzzy sets *)
   'v_small/small/l_small/l_large/large/v_large',         (* 6 fuzzy sets *)
   'v_small/small/l_small/medium/l_large/large/v_large'); (* 7 fuzzy sets *)


type

  (* This is a triangular fuzzy set with 3 parameters, a: left bottom      *)
  (* b: top, c: right base; mu(a) = mu(c) = 0, mu(b) = 1. The triangle can *)
  (* be left shouldered mu(a) = 1 or right shouldered mu(c) = 1.           *)
  TriangleFSet = record
                  Name             : FSNStr;
                  a,b,c            : real;
                  left_shouldered  : boolean;
                  right_shouldered : boolean;
                 end;

  (* Partioning of a variable: a number of triangular fuzzy sets *)
  PPartition = ^Partition;
  Partition = record
               VName : VNStr;
               count : integer;
               Fsets : array[0..maxfs] of TriangleFSet;
              end;

  (* weighted edge *)
  Edge = record
          node  : integer;
          weight: real;
         end;

  (* A range *)
  Range = record
           min: real;
           max: real;
          end;

  (* A pair of pattern and rule indeces *)

  RuleCreator = record
      Pattern : LongInt;
      Rule    : integer;
  end;

  (* description of a linguistic term in text form *)
  LingTerm = record
     Variable: VNStr;
     Value   : VNStr;
  end;

  (* A description of a rule of the network in text form *)
  PRuleDescriptor = ^RuleDescriptor;
  RuleDescriptor = record
     Name       : VNStr;
     Antecedent : array[1..maxunits] of LingTerm;
     Conclusion : VNStr;
  end;

  (* One Partioning for each variable *)
  Farray          = array[1..maxunits] of Partition;

  (* Array to store indeces of fuzzy sets *)
  ByteVector      = array[1..RuleLimit] of byte;

  (* Connection matrix, entries are indeces and point to fuzzy sets *)
  WeightMatrix    = array[1..maxunits] of ByteVector;

  (* Conclusions: number of output unit and connection weight *)
  WeightVector    = array[1..RuleLimit] of Edge;

  (* Array for activations and weights *)
  UnitVector      = array[1..maxunits] of real;
  PUnitVector     = ^UnitVector;

  (* Array for activations of rule units *)
  RuleVector      = array[1..RuleLimit] of real;

  (* Array for features of patterns *)
  FeatureVector   = array[1..maxunits] of real;

  (* Array of to show which input unit needs a new partitioning *)
  FlagVector      = array[1..maxunits] of boolean;

  (* A buffer for error values *)
  TBuffer          = array[0..maxbuf-1,1..3] of real;

  (* Array to store which pattern caused what rule to be created *)
  RuleCreatorArray = array [0..RuleLimit] of RuleCreator;

  (* Array to store membership values for all fuzzy sets of all features *)
  MBFArray = array [1..maxunits,1..maxfs] of real;

  RRanges = record
             Min, Max: UnitVector;
            end;

  RuleLearningType = (Simple, Best, BestPerClass);

  PRLearnParam = ^TRLearnParam;
  TRLearnParam = record
      EndEpoch     : LongInt;
      ErrEpochs    : LongInt;
      UpdateRate   : Word;
      Misclass_OK  : Word;
      StopError    : Real;
      s, sa, sb, sc: real;
      RuleLearning : RuleLearningType;
      StopCheck    : Boolean;
      TextDisplay  : Boolean;
      free_w_learn : Boolean;
      w_learn      : Boolean;
      intersect05  : Boolean;
      asymmetric   : Boolean;
      nopassing    : Boolean;
      single_step  : Boolean;
      ReLearnRules : Boolean;
  end;

  PRControlParam = ^TRControlParam;
  TRControlParam = record
      EndEpoch     : LongInt;
      ErrEpochs    : LongInt;
      UpdateRate   : Word;
      Misclass_OK  : Word;
      StopError    : Real;
      StopCheck    : Boolean;
      TextDisplay  : Boolean;
      single_step  : Boolean;
  end;

  PPartCollection = ^TPartCollection;
  TPartCollection = object (TCollection)
   procedure FreeItem (Item: pointer); virtual;
   procedure Kill; virtual;
  end;

  PMyStrCollection = ^TMyStrCollection;
  TMyStrCollection = object (TCollection)
   procedure FreeItem (Item: pointer); virtual;
   procedure Kill; virtual;
  end;

  PAntStrCollection = ^TAntStrCollection;
  TAntStrCollection = object (TCollection)
   procedure FreeItem (Item: pointer); virtual;
   procedure Kill; virtual;
  end;

  PRuleStrCollection = ^TRuleStrCollection;
  TRuleStrCollection = object (TCollection)
   procedure FreeItem (Item: pointer); virtual;
   procedure Kill; virtual;
  end;

  PVectorCollection = ^TVectorCollection;
  TVectorCollection = object (TCollection)
   procedure FreeItem (Item: pointer); virtual;
  end;

var (* for global debugging *)
 RCIndex : RuleCreatorArray;

var (* global path names *)
 PathToBGI : string;

function skip_comments (var ParamFile: text): string;
function Filename (var ParamFile: Text; ReadFile: boolean;
                   FN: FNameStr): boolean;
procedure OpenErrorFile (var EFile: Text; NewFile: Boolean);
procedure WriteErrorFile (var EFile: Text; var EBuf: TBuffer; Count: integer);
procedure WriteRuleCreators (var EFile: Text; var RC: RuleCreatorArray;
                                 Count: integer);
procedure CloseErrorFile (var EFile: Text);
function sgn(x: real): integer;
procedure GetFName (row, col: integer; var Name: FSNStr);
procedure WriteFileHeader (var ParamFile: Text; Comment: string);
function CalcHelpName: PathStr;

implementation
(*-------------------------------------------------------------------------*)
(* This function skips comments and blank lines, and gets the next         *)
(* command in a network or pattern file                                    *)
(*-------------------------------------------------------------------------*)
function skip_comments (var ParamFile: text): string;
 var
  line : string;
  i,p  : integer;
 begin
  repeat
   {$I-}
   readln (ParamFile, line);
   {$I+}
   if IOResult <> 0 then line := 'END';
   while line[1] = ' ' do
    line := copy (line, 1, length(line));
  until (line[1] <> '%') and (line <> '');
  p := pos(line,' ');
  if p <> 0
   then delete (line, p , length (line));
  for i := 1 to length(line) do
   line[i] := upcase(line[i]);
  skip_comments := line;
 end; (* skip_comments *)

(*-------------------------------------------------------------------------*)
(* This function checks if a file can be opened for reading or writing. It *)
(* gets a filename in FN, and returns a filehandle in the first parameter, *)
(* and the result TRUE, if the file can be created/opened.                 *)
(*-------------------------------------------------------------------------*)

function Filename (var ParamFile: Text; ReadFile: boolean;
                   FN: FNameStr): boolean;
 var
  OK      : boolean;
  Error   : boolean;
  MSGParam: PString;

 begin
   OK := False;
   MSGParam := @FN;
   Error := False;
   if FN = '' then Error := True;
   Assign (ParamFile, FN);
   If ReadFile
    then begin
     if not Error
       then begin
        {$I-}
        Reset (ParamFile);
        {$I+}
       end;
     if (IOResult <> 0) or Error
       then MessageBox ('The file %s does not exist!', @MSGParam,
                         mfError or mfOKButton)
       else begin
        OK := True;
        close (ParamFile);
       end;
    end
    else begin
     if not Error
       then begin
        {$I-}
        Rewrite (ParamFile);
        {$I+}
       end;
     if (IOResult <> 0) or Error
       then MessageBox ('The file %s cannot be created!', @MSGParam,
                         mfError or mfOKButton)
       else begin
        OK := True;
        close (Paramfile);
       end;
    end;
  Filename := Ok;
 end; (* Filename *)

(*-------------------------------------------------------------------------*)

procedure OpenErrorFile (var EFile: Text; NewFile: Boolean);
 begin
  Assign (EFile, 'NEFCLASS.PRO');
  if NewFile
    then Rewrite (EFile)
    else Append (EFile);
 end;

(*-------------------------------------------------------------------------*)

procedure WriteErrorFile (var EFile: Text; var EBuf: TBuffer; Count: integer);
 var
  i : integer;
 begin
  i := 0;
  while i < Count do
   begin
    writeln (EFile, EBuf[i,1]:7:0, ' ', EBuf[i,2]:10:6, ' ', EBuf[i,3]:4:0);
    inc (i);
   end;
 end;

(*-------------------------------------------------------------------------*)

procedure WriteRuleCreators (var EFile: Text; var RC: RuleCreatorArray;
                                 Count: integer);
 var
  i : integer;
 begin
  i := 0;
  writeln (EFile, '% Patterns that have created rules');
  writeln (EFile);

  for i := 1 to Count do
   writeln (EFile, 'Pattern ', RC[i].Pattern:10, ' created rule ',
            RC[i].Rule:10);
  writeln (EFile);
 end;

(*-------------------------------------------------------------------------*)

procedure CloseErrorFile (var EFile: Text);
 begin
  close (EFile);
 end;

(*-------------------------------------------------------------------------*)

function sgn(x: real): integer;
 begin
  if x < 0
    then sgn := -1
    else if x > 0
           then sgn := 1
           else sgn := 0;
 end; (* sgn *)

(*-------------------------------------------------------------------------*)

procedure GetFName (row, col: integer; var Name: FSNStr);
 var
  rest : FSNCStr;
  i, p : integer;
 begin
  if (row >= 0) and (row <= 7) and (col >= 1) and ((col <= row) or (row = 0))
    then begin
     rest := FNames[row];
     for i := 1 to col-1 do
      rest := Copy (rest, Pos ('/', rest) + 1,
                    Length(rest));
     p := Pos('/', rest);
     if p = 0
       then Name := rest
       else Name := Copy (rest, 1, p-1);
    end
    else Name := '';
 end; (* GetFName *)

(*-------------------------------------------------------------------------*)

procedure WriteFileHeader (var ParamFile: Text; Comment: string);
 var
  i, j   : integer;
  ls, rs : integer;
  h,m,s,ss: word; (* to get the time *)
  min     : string[2];
  y,mo,d,w : word; (* to get the date *)

 begin

  GetDate (y,mo,d,w); GetTime (h,m,s,ss);
  str (m:2, min);
  if m < 10 then min[1] := '0';

  writeln (ParamFile, '% NEFCLASS ', VersionString,
                      ', (c) TU Braunschweig, 1994');
  writeln (ParamFile, '% This file was created on ', mo,'/',d,'/',y, ' at ',
                      h,':',min, '.');
  writeln (ParamFile, '%');
  writeln (ParamFile, '% ', Comment);
  writeln (ParamFile);

 end; (* WriteFileHeader *)

(*-------------------------------------------------------------------------*)


function CalcHelpName: PathStr;
var
  EXEName : PathStr;
  Dir     : DirStr;
  Name    : NameStr;
  Ext     : ExtStr;
begin
  if Lo(DosVersion) >= 3 then EXEName := ParamStr(0)
  else EXEName := FSearch('NFCL_TV.EXE', GetEnv('PATH'));
  FSplit(EXEName, Dir, Name, Ext);
  if Dir[Length(Dir)] = '\' then Dec(Dir[0]);
  CalcHelpName := FSearch('NFCLHELP.HLP', Dir);
end;

(* ==================== Methods of TPartCollection ======================= *)

 procedure TPartCollection.FreeItem (Item: Pointer);
  begin
   if Item <> nil then Dispose (PPartition(Item));
  end;

 procedure TPartCollection.Kill;
  begin
   Dispose(PObject(@Self), Done);
  end;

(* ==================== Methods of TMyStrCollection ====================== *)

 procedure TMyStrCollection.FreeItem (Item: Pointer);
  begin
   if Item <> nil then Dispose (PVNStr(Item));
  end;

 procedure TMyStrCollection.Kill;
  begin
   Dispose(PObject(@Self), Done);
  end;

(* ==================== Methods of TAntStrCollection ===================== *)

 procedure TAntStrCollection.FreeItem (Item: Pointer);
  begin
   if Item <> nil then Dispose (PAntStr(Item));
  end;

 procedure TAntStrCollection.Kill;
  begin
   Dispose(PObject(@Self), Done);
  end;

(* ==================== Methods of TRuleStrCollection ==================== *)

 procedure TRuleStrCollection.FreeItem (Item: Pointer);
  begin
   if Item <> nil then Dispose (PRuleDescriptor(Item));
  end;

 procedure TRuleStrCollection.Kill;
  begin
   Dispose(PObject(@Self), Done);
  end;

(* ==================== Methods of TRuleStrCollection ==================== *)

 procedure TVectorCollection.FreeItem (Item: Pointer);
  begin
   if Item <> nil then Dispose (PUnitVector(Item));
  end;

begin
  PathToBGI := GetEnv ('BGIPATH');
end.